from .__version__ import __version__
from .density import Density
from .preprocessor import Preprocessor
from .structure import Structure
from .matching_optimization import FitRefinement
from . import extensions



#!python3
""" Apply tme.preprocessor.Preprocessor methods to an input file based
    on a provided yaml configuration obtaiend from preprocessor_gui.py.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import yaml
import argparse
import textwrap
from tme import Preprocessor, Density


def parse_args():
    parser = argparse.ArgumentParser(
        description=textwrap.dedent(
            """
        Apply preprocessing to an input file based on a provided YAML configuration.

        Expected YAML file format:
        ```yaml
        <method_name>:
            <parameter1>: <value1>
            <parameter2>: <value2>
            ...
        ```
        """
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--input_file",
        type=str,
        required=True,
        help="Path to the input data file in CCP4/MRC format.",
    )
    parser.add_argument(
        "-y",
        "--yaml_file",
        type=str,
        required=True,
        help="Path to the YAML configuration file.",
    )
    parser.add_argument(
        "-o",
        "--output_file",
        type=str,
        required=True,
        help="Path to output file in CPP4/MRC format..",
    )
    parser.add_argument(
        "--compress", action="store_true", help="Compress the output file using gzip."
    )

    args = parser.parse_args()

    return args


def main():
    args = parse_args()
    with open(args.yaml_file, "r") as f:
        preprocess_settings = yaml.safe_load(f)

    if len(preprocess_settings) > 1:
        raise NotImplementedError(
            "Multiple preprocessing methods specified. "
            "The script currently supports one method at a time."
        )

    method_name = list(preprocess_settings.keys())[0]
    if not hasattr(Preprocessor, method_name):
        raise ValueError(f"Method {method_name} does not exist in Preprocessor.")

    density = Density.from_file(args.input_file)
    output = density.empty

    method_params = preprocess_settings[method_name]
    preprocessor = Preprocessor()
    method = getattr(preprocessor, method_name, None)
    if not method:
        raise ValueError(
            f"{method} does not exist in dge.preprocessor.Preprocessor class."
        )

    output.data = method(template=density.data, **method_params)
    output.to_file(args.output_file, gzip=args.compress)


if __name__ == "__main__":
    main()



""" Utility functions for template matching.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import os
import traceback
import pickle
from shutil import move
from tempfile import mkstemp
from itertools import product
from typing import Tuple, Dict, Callable
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from numpy.typing import NDArray
from scipy.spatial import ConvexHull
from scipy.spatial.transform import Rotation

from .helpers import quaternion_to_rotation_matrix, load_quaternions_by_angle
from .extensions import max_euclidean_distance
from .matching_memory import estimate_ram_usage


def handle_traceback(last_type, last_value, last_traceback):
    """
    Handle sys.exc_info().

    Parameters
    ----------
    last_type : type
        The type of the last exception.
    last_value :
        The value of the last exception.
    last_traceback : traceback
        The traceback object encapsulating the call stack at the point
        where the exception originally occurred.

    Raises
    ------
    Exception
        Re-raises the last exception.
    """
    if last_type is None:
        return None
    traceback.print_tb(last_traceback)
    raise Exception(last_value)
    # raise last_type(last_value)


def generate_tempfile_name(suffix=None):
    """
    Returns the path to a potential temporary file location. If the environment
    variable TME_TMPDIR is defined, the temporary file will be created there.
    Otherwise the default tmp directory will be used.

    Parameters
    ----------
    suffix : str, optional
        File suffix. By default the file has no suffix.

    Returns
    -------
    str
        The generated filename
    """
    tmp_dir = os.environ.get("TMPDIR", None)
    _, filename = mkstemp(suffix=suffix, dir=tmp_dir)
    return filename


def array_to_memmap(arr: NDArray, filename: str = None) -> str:
    """
    Converts a numpy array to a np.memmap.

    Parameters
    ----------
    arr : np.ndarray
        The numpy array to be converted.
    filename : str, optional
        Desired filename for the memmap. If not provided, a temporary
        file will be created.

    Notes
    -----
        If the environment variable TME_TMPDIR is defined, the temporary
        file will be created there. Otherwise the default tmp directory
        will be used.

    Returns
    -------
    str
        The filename where the memmap was written to.
    """
    if filename is None:
        filename = generate_tempfile_name()

    shape, dtype = arr.shape, arr.dtype
    arr_memmap = np.memmap(filename, mode="w+", dtype=dtype, shape=shape)

    arr_memmap[:] = arr[:]
    arr_memmap.flush()

    return filename


def memmap_to_array(arr: NDArray) -> NDArray:
    """
    Converts a np.memmap into an numpy array.

    Parameters
    ----------
    arr : np.memmap
        The numpy array to be converted.

    Returns
    -------
    np.ndarray
        The converted array.
    """
    if type(arr) == np.memmap:
        memmap_filepath = arr.filename
        arr = np.array(arr)
        os.remove(memmap_filepath)
    return arr


def close_memmap(arr: np.ndarray) -> None:
    """
    Remove the file associated with a numpy memmap array.

    Parameters
    ----------
    arr : np.ndarray
        The numpy array which might be a memmap.
    """
    try:
        os.remove(arr.filename)
        # arr._mmap.close()
    except Exception:
        pass


def write_pickle(data: object, filename: str) -> None:
    """
    Serialize and write data to a file invalidating the input data in
    the process. This function  uses type-specific serialization for
    certain objects, such as np.memmap, for optimized storage. Other
    objects are serialized using standard pickle.

    Parameters
    ----------
    data : iterable or object
        The data to be serialized.
    filename : str
        The name of the file where the serialized data will be written.

    See Also
    --------
    :py:meth:`load_pickle`
    """
    if type(data) not in (list, tuple):
        data = (data,)

    dirname = os.path.dirname(filename)
    with open(filename, "wb") as ofile, ThreadPoolExecutor() as executor:
        for i in range(len(data)):
            futures = []
            item = data[i]
            if isinstance(item, np.memmap):
                _, new_filename = mkstemp(suffix=".mm", dir=dirname)
                new_item = ("np.memmap", item.shape, item.dtype, new_filename)
                futures.append(executor.submit(move, item.filename, new_filename))
                item = new_item
            pickle.dump(item, ofile)
        for future in futures:
            future.result()


def load_pickle(filename: str) -> object:
    """
    Load and deserialize data written by :py:meth:`write_pickle`.

    Parameters
    ----------
    filename : str
        The name of the file to read and deserialize data from.

    Returns
    -------
    object or iterable
        The deserialized data.

    See Also
    --------
    :py:meth:`write_pickle`
    """

    def _load_pickle(file_handle):
        try:
            while True:
                yield pickle.load(file_handle)
        except EOFError:
            pass

    def _is_pickle_memmap(data):
        ret = False
        if type(data[0]) == str:
            if data[0] == "np.memmap":
                ret = True
        return ret

    items = []
    with open(filename, "rb") as ifile:
        for data in _load_pickle(ifile):
            if isinstance(data, tuple):
                if _is_pickle_memmap(data):
                    _, shape, dtype, filename = data
                    data = np.memmap(filename, shape=shape, dtype=dtype)
            items.append(data)
    return items[0] if len(items) == 1 else items


def compute_parallelization_schedule(
    shape1: NDArray,
    shape2: NDArray,
    max_cores: int,
    max_ram: int,
    matching_method: str,
    backend: str = None,
    split_only_outer: bool = False,
    shape1_padding: NDArray = None,
    analyzer_method: str = None,
    max_splits: int = 256,
    float_nbytes: int = 4,
    complex_nbytes: int = 8,
    integer_nbytes: int = 4,
) -> Tuple[Dict, int, int]:
    """
    Computes a parallelization schedule for a given computation.

    This function estimates the amount of memory that would be used by a computation
    and breaks down the computation into smaller parts that can be executed in parallel
    without exceeding the specified limits on the number of cores and memory.

    Parameters
    ----------
    shape1 : NDArray
        The shape of the first input tensor.
    shape1_padding : NDArray, optional
        Padding for shape1 used for each split. None by defauly
    shape2 : NDArray
        The shape of the second input tensor.
    max_cores : int
        The maximum number of cores that can be used.
    max_ram : int
        The maximum amount of memory that can be used.
    matching_method : str
        The metric used for scoring the computations.
    backend : str, optional
        Backend used for computations.
    split_only_outer : bool, optional
        Whether only outer splits sould be considered.
    analyzer_method : str
        The method used for score analysis.
    max_splits : int, optional
        The maximum number of parts that the computation can be split into,
        by default 256.
    float_nbytes : int
        Number of bytes of the used float, e.g. 4 for float32.
    complex_nbytes : int
        Number of bytes of the used complex, e.g. 8 for complex64.
    integer_nbytes : int
        Number of bytes of the used integer, e.g. 4 for int32.

    Notes
    -----
        This function assumes that no residual memory remains after each split,
        which not always holds true, e.g. when using
        :py:class:`tme.analyzer.MaxScoreOverRotations`.

    Returns
    -------
    dict
        The optimal splits for each axis of the first input tensor.
    int
        The number of outer jobs.
    int
        The number of inner jobs per outer job.
    """
    shape1, shape2 = np.array(shape1), np.array(shape2)
    if shape1_padding is None:
        shape1_padding = np.zeros_like(shape1)
    core_assignments = []
    for i in range(1, int(max_cores**0.5) + 1):
        if max_cores % i == 0:
            core_assignments.append((i, max_cores // i))
            core_assignments.append((max_cores // i, i))

    if split_only_outer:
        core_assignments = [(1, max_cores)]

    possible_params, split_axis = [], np.argmax(shape1)
    split_factor, n_splits = [1 for _ in range(len(shape1))], 0
    while n_splits <= max_splits:
        splits = {k: split_factor[k] for k in range(len(split_factor))}
        array_slices = split_numpy_array_slices(shape=shape1, splits=splits)
        array_widths = [
            tuple(x.stop - x.start for x in split) for split in array_slices
        ]
        n_splits = np.prod(list(splits.values()))

        for inner_cores, outer_cores in core_assignments:
            if outer_cores > n_splits:
                continue
            ram_usage = [
                estimate_ram_usage(
                    shape1=np.add(shp, shape1_padding),
                    shape2=shape2,
                    matching_method=matching_method,
                    analyzer_method=analyzer_method,
                    backend=backend,
                    ncores=inner_cores,
                    float_nbytes=float_nbytes,
                    complex_nbytes=complex_nbytes,
                    integer_nbytes=integer_nbytes,
                )
                for shp in array_widths
            ]
            max_usage = 0
            for i in range(0, len(ram_usage), outer_cores):
                usage = np.sum(ram_usage[i : (i + outer_cores)])
                if usage > max_usage:
                    max_usage = usage

            inits = n_splits // outer_cores
            if max_usage < max_ram:
                possible_params.append(
                    (*split_factor, outer_cores, inner_cores, n_splits, inits)
                )
        split_factor[split_axis] += 1
        split_axis += 1
        if split_axis == shape1.size:
            split_axis = 0

    possible_params = np.array(possible_params)
    if not len(possible_params):
        print(
            "No suitable assignment found. Consider increasing "
            "max_ram or decrease max_cores."
        )
        return None, None

    init = possible_params.shape[1] - 1
    possible_params = possible_params[
        np.lexsort((possible_params[:, init], possible_params[:, (init - 1)]))
    ]
    splits = {k: possible_params[0, k] for k in range(shape1.size)}
    core_assignment = (
        possible_params[0, shape1.size],
        possible_params[0, (shape1.size + 1)],
    )

    return splits, core_assignment


def centered(arr: NDArray, newshape: Tuple[int]) -> NDArray:
    """
    Extract the centered portion of an array based on a new shape.

    Parameters
    ----------
    arr : NDArray
        Input array.
    newshape : tuple
        Desired shape for the central portion.

    Returns
    -------
    NDArray
        Central portion of the array with shape `newshape`.

    References
    ----------
    .. [1] https://github.com/scipy/scipy/blob/v1.11.2/scipy/signal/_signaltools.py#L388
    """
    new_shape = np.asarray(newshape)
    current_shape = np.array(arr.shape)
    starts = (current_shape - new_shape) // 2
    stops = starts + newshape
    box = tuple(slice(start, stop) for start, stop in zip(starts, stops))
    return arr[box]


def centered_mask(arr: NDArray, newshape: Tuple[int]) -> NDArray:
    """
    Mask the centered portion of an array based on a new shape.

    Parameters
    ----------
    arr : NDArray
        Input array.
    newshape : tuple
        Desired shape for the mask.

    Returns
    -------
    NDArray
        Array with central portion unmasked and the rest set to 0.
    """
    new_shape = np.asarray(newshape)
    current_shape = np.array(arr.shape)
    starts = (current_shape - new_shape) // 2
    stops = starts + newshape
    box = tuple(slice(start, stop) for start, stop in zip(starts, stops))
    mask = np.zeros_like(arr)
    mask[box] = 1
    arr *= mask
    return arr


def apply_convolution_mode(
    arr: NDArray,
    convolution_mode: str,
    s1: Tuple[int],
    s2: Tuple[int],
    mask_output: bool = False,
) -> NDArray:
    """
    Applies convolution_mode to arr.

    Parameters
    ----------
    arr : NDArray
        Numpy array containing convolution result of arrays with shape s1 and s2.
    convolution_mode : str
        Analogous to mode in ``scipy.signal.convolve``:

        +---------+----------------------------------------------------------+
        | 'full'  | returns full template matching result of the inputs.     |
        +---------+----------------------------------------------------------+
        | 'valid' | returns elements that do not rely on zero-padding..      |
        +---------+----------------------------------------------------------+
        | 'same'  | output is the same size as s1.                           |
        +---------+----------------------------------------------------------+
    s1 : tuple
        Tuple of integers corresponding to shape of convolution array 1.
    s2 : tuple
        Tuple of integers corresponding to shape of convolution array 2.
    mask_output : bool, optional
        Whether to mask values outside of convolution_mode rather than
        removing them. Defaults to False.

    Returns
    -------
    NDArray
        The numpy array after applying the convolution mode.

    References
    ----------
    .. [1] https://github.com/scipy/scipy/blob/v1.11.2/scipy/signal/_signaltools.py#L519
    """
    # This removes padding to next fast fourier length
    arr = arr[tuple(slice(s1[i] + s2[i] - 1) for i in range(len(s1)))]

    if convolution_mode not in ("full", "same", "valid"):
        raise ValueError("Supported convolution_mode are 'full', 'same' and 'valid'.")

    func = centered_mask if mask_output else centered
    if convolution_mode == "full":
        return arr
    elif convolution_mode == "same":
        return func(arr, s1)
    elif convolution_mode == "valid":
        valid_shape = [s1[i] - s2[i] + s2[i] % 2 for i in range(arr.ndim)]
        return func(arr, valid_shape)


def compute_full_convolution_index(
    outer_shape: Tuple[int],
    inner_shape: Tuple[int],
    outer_split: Tuple[slice],
    inner_split: Tuple[slice],
) -> Tuple[slice]:
    """
    Computes the position of the convolution of pieces in the full convolution.

    Parameters
    ----------
    outer_shape : tuple
        Tuple of integers corresponding to the shape of the outer array.
    inner_shape : tuple
        Tuple of integers corresponding to the shape of the inner array.
    outer_split : tuple
        Tuple of slices used to split outer array
        (see :py:meth:`split_numpy_array_slices`).
    inner_split : tuple
        Tuple of slices used to split inner array
        (see :py:meth:`split_numpy_array_slices`).

    Returns
    -------
    tuple
        Tuple of slices corresponding to the position of the given convolution
        in the full convolution.
    """
    outer_shape = np.asarray(outer_shape)
    inner_shape = np.asarray(inner_shape)

    outer_width = np.array([outer.stop - outer.start for outer in outer_split])
    inner_width = np.array([inner.stop - inner.start for inner in inner_split])
    convolution_shape = outer_width + inner_width - 1

    end_inner = np.array([inner.stop for inner in inner_split]).astype(int)
    start_outer = np.array([outer.start for outer in outer_split]).astype(int)

    offsets = start_outer + inner_shape - end_inner

    score_slice = tuple(
        (slice(offset, offset + shape))
        for offset, shape in zip(offsets, convolution_shape)
    )

    return score_slice


def split_numpy_array_slices(
    shape: NDArray, splits: Dict, margin: NDArray = None
) -> Tuple[slice]:
    """
    Returns a tuple of slices to subset a numpy array into pieces along multiple axes.

    Parameters
    ----------
    shape : NDArray
        Shape of the array to split.
    splits : dict
        A dictionary where the keys are the axis numbers and the values
        are the number of splits along that axis.
    margin : NDArray, optional
        Padding on the left hand side of the array.

    Returns
    -------
    tuple
        A tuple of slices, where each slice corresponds to a split along an axis.
    """
    ndim = len(shape)
    if margin is None:
        margin = np.zeros(ndim, dtype=int)
    splits = {k: max(splits.get(k, 0), 1) for k in range(ndim)}
    new_shape = np.divide(shape, [splits.get(i, 1) for i in range(ndim)]).astype(int)

    slice_list = [
        tuple(
            (slice(max((n_splits * length) - margin[axis], 0), (n_splits + 1) * length))
            if n_splits < splits.get(axis, 1) - 1
            else (slice(max((n_splits * length) - margin[axis], 0), shape[axis]))
            for n_splits in range(splits.get(axis, 1))
        )
        for length, axis in zip(new_shape, splits.keys())
    ]

    splits = tuple(product(*slice_list))

    return splits


def get_rotation_matrices(
    angular_sampling: float, dim: int = 3, use_optimized_set: bool = True
) -> NDArray:
    """
    Returns rotation matrices in format k x dim x dim, where k is determined
    by ``angular_sampling``.

    Parameters
    ----------
    angular_sampling : float
        The angle in degrees used for the generation of rotation matrices.
    dim : int, optional
        Dimension of the rotation matrices.
    use_optimized_set : bool, optional
        Whether to use pre-computed rotational sets with more optimal sampling.
        Currently only available when dim=3.

    Notes
    -----
        For the case of dim = 3 optimized rotational sets are used, otherwise
        QR-decomposition.

    Returns
    -------
    NDArray
        Array of shape (k, dim, dim) containing k rotation matrices.
    """
    if dim == 3 and use_optimized_set:
        quaternions, *_ = load_quaternions_by_angle(angular_sampling)
        ret = quaternion_to_rotation_matrix(quaternions)
    else:
        num_rotations = dim * (dim - 1) // 2
        k = int((360 / angular_sampling) ** num_rotations)
        As = np.random.randn(k, dim, dim)
        ret, _ = np.linalg.qr(As)
        dets = np.linalg.det(ret)
        neg_dets = dets < 0
        ret[neg_dets, :, -1] *= -1
    return ret


def minimum_enclosing_box(
    coordinates: NDArray,
    margin: NDArray = None,
    use_geometric_center: bool = False,
) -> Tuple[int]:
    """
    Computes the minimal enclosing box around coordinates with margin.

    Parameters
    ----------
    coordinates : NDArray
        Coordinates of which the enclosing box should be computed. The shape
        of this array should be [d, n] with d dimensions and n coordinates.
    margin : NDArray, optional
        Box margin. Defaults to None.
    use_geometric_center : bool, optional
        Whether the box should accommodate the geometric or the coordinate
        center. Defaults to False.

    Returns
    -------
    tuple
        Integers corresponding to the minimum enclosing box shape.
    """
    point_cloud = np.asarray(coordinates)
    dim = point_cloud.shape[0]
    point_cloud = point_cloud - point_cloud.min(axis=1)[:, None]

    margin = np.zeros(dim) if margin is None else margin
    margin = np.asarray(margin).astype(int)

    norm_cloud = point_cloud - point_cloud.mean(axis=1)[:, None]
    # Adding one avoids clipping during scipy.ndimage.affine_transform
    shape = np.repeat(
        np.ceil(2 * np.linalg.norm(norm_cloud, axis=0).max()) + 1, dim
    ).astype(int)
    if use_geometric_center:
        hull = ConvexHull(point_cloud.T)
        distance, _ = max_euclidean_distance(point_cloud[:, hull.vertices].T)
        distance += np.linalg.norm(np.ones(dim))
        shape = np.repeat(np.rint(distance).astype(int), dim)

    return shape


def crop_input(
    target: "Density",
    template: "Density",
    target_mask: "Density" = None,
    template_mask: "Density" = None,
    map_cutoff: float = 0,
    template_cutoff: float = 0,
) -> Tuple[int]:
    """
    Crop target and template maps for efficient fitting. Input densities
    are cropped in place.

    Parameters
    ----------
    target : Density
        Target to be fitted on.
    template : Density
        Template to fit onto the target.
    target_mask : Density, optional
        Path to mask of target. Will be croppped like target.
    template_mask : Density, optional
        Path to mask of template. Will be cropped like template.
    map_cutoff : float, optional
        Cutoff value for trimming the target Density. Default is 0.
    map_cutoff : float, optional
        Cutoff value for trimming the template Density. Default is 0.

    Returns
    -------
    Tuple[int]
        Tuple containing reference fit index
    """
    convolution_shape_init = np.add(target.shape, template.shape) - 1
    # If target and template are aligned, fitting should return this index
    reference_fit = np.subtract(template.shape, 1)

    target_box = tuple(slice(0, x) for x in target.shape)
    if map_cutoff is not None:
        target_box = target.trim_box(cutoff=map_cutoff)

    target_mask_box = target_box
    if target_mask is not None and map_cutoff is not None:
        target_mask_box = target_mask.trim_box(cutoff=map_cutoff)
    target_box = tuple(
        slice(min(arr.start, mask.start), max(arr.stop, mask.stop))
        for arr, mask in zip(target_box, target_mask_box)
    )

    template_box = tuple(slice(0, x) for x in template.shape)
    if template_cutoff is not None:
        template_box = template.trim_box(cutoff=template_cutoff)

    template_mask_box = template_box
    if template_mask is not None and template_cutoff is not None:
        template_mask_box = template_mask.trim_box(cutoff=template_cutoff)
    template_box = tuple(
        slice(min(arr.start, mask.start), max(arr.stop, mask.stop))
        for arr, mask in zip(template_box, template_mask_box)
    )

    cut_right = np.array(
        [shape - x.stop for shape, x in zip(template.shape, template_box)]
    )
    cut_left = np.array([x.start for x in target_box])

    origin_difference = np.divide(target.origin - template.origin, target.sampling_rate)
    origin_difference = origin_difference.astype(int)

    target.adjust_box(target_box)
    template.adjust_box(template_box)

    if target_mask is not None:
        target_mask.adjust_box(target_box)
    if template_mask is not None:
        template_mask.adjust_box(template_box)

    reference_fit -= cut_right + cut_left + origin_difference

    convolution_shape = np.array(target.shape)
    convolution_shape += np.array(template.shape) - 1

    print(f"Cropped volume of target is: {target.shape}")
    print(f"Cropped volume of template is: {template.shape}")
    saving = 1 - (np.prod(convolution_shape)) / np.prod(convolution_shape_init)
    saving *= 100

    print(
        "Cropping changed array size from "
        f"{round(4*np.prod(convolution_shape_init)/1e6, 3)} MB "
        f"to {round(4*np.prod(convolution_shape)/1e6, 3)} MB "
        f"({'-' if saving > 0 else ''}{abs(round(saving, 2))}%)"
    )
    return reference_fit


def euler_to_rotationmatrix(angles: Tuple[float]) -> NDArray:
    """
    Convert Euler angles to a rotation matrix.

    Parameters
    ----------
    angles : tuple
        A tuple representing the Euler angles in degrees.

    Returns
    -------
    NDArray
        The generated rotation matrix.
    """
    if len(angles) == 1:
        angles = (angles, 0, 0)
    rotation_matrix = (
        Rotation.from_euler("zyx", angles, degrees=True).as_matrix().astype(np.float32)
    )
    return rotation_matrix


def euler_from_rotationmatrix(rotation_matrix: NDArray) -> Tuple:
    """
    Convert a rotation matrix to euler angles.

    Parameters
    ----------
    rotation_matrix : NDArray
        A 2 x 2 or 3 x 3 rotation matrix in z y x form.

    Returns
    -------
    Tuple
        The generate euler angles in degrees
    """
    if rotation_matrix.shape[0] == 2:
        temp_matrix = np.eye(3)
        temp_matrix[:2, :2] = rotation_matrix
        rotation_matrix = temp_matrix
    euler_angles = (
        Rotation.from_matrix(rotation_matrix)
        .as_euler("zyx", degrees=True)
        .astype(np.float32)
    )
    return euler_angles


def rigid_transform(
    coordinates: NDArray,
    rotation_matrix: NDArray,
    out: NDArray,
    translation: NDArray,
    use_geometric_center: bool = False,
    coordinates_mask: NDArray = None,
    out_mask: NDArray = None,
    center: NDArray = None,
) -> None:
    """
    Apply a rigid transformation (rotation and translation) to given coordinates.

    Parameters
    ----------
    coordinates : NDArray
        An array representing the coordinates to be transformed [d x N].
    rotation_matrix : NDArray
        The rotation matrix to be applied [d x d].
    translation : NDArray
        The translation vector to be applied [d].
    out : NDArray
        The output array to store the transformed coordinates.
    coordinates_mask : NDArray, optional
        An array representing the mask for the coordinates [d x T].
    out_mask : NDArray, optional
        The output array to store the transformed coordinates mask.
    use_geometric_center : bool, optional
        Whether to use geometric or coordinate center.

    Returns
    -------
    None
    """
    coordinate_dtype = coordinates.dtype
    center = coordinates.mean(axis=1) if center is None else center
    if not use_geometric_center:
        coordinates = coordinates - center[:, None]

    np.matmul(rotation_matrix, coordinates, out=out)
    if use_geometric_center:
        axis_max, axis_min = out.max(axis=1), out.min(axis=1)
        axis_difference = axis_max - axis_min
        translation = np.add(translation, center - axis_max + (axis_difference // 2))
    else:
        translation = np.add(translation, np.subtract(center, out.mean(axis=1)))

    out += translation[:, None]
    if coordinates_mask is not None and out_mask is not None:
        if not use_geometric_center:
            coordinates_mask = coordinates_mask - center[:, None]
        np.matmul(rotation_matrix, coordinates_mask, out=out_mask)
        out_mask += translation[:, None]

    if not use_geometric_center and coordinate_dtype != out.dtype:
        np.subtract(out.mean(axis=1), out.astype(int).mean(axis=1), out=translation)
        out += translation[:, None]


def _format_string(string: str) -> str:
    """
    Formats a string by adding quotation marks if it contains white spaces.

    Parameters
    ----------
    string : str
        Input string to be formatted.

    Returns
    -------
    str
        Formatted string with added quotation marks if needed.
    """
    if " " in string:
        return f"'{string}'"
    # Occurs e.g. for C1' atoms. The trailing whitespace is necessary.
    if string.count("'") == 1:
        return f'"{string}"'
    return string


def _format_mmcif_colunns(subdict: Dict) -> Dict:
    """
    Formats the columns of a mmcif dictionary.

    Parameters
    ----------
    subdict : dict
        Input dictionary where each key corresponds to a column and the
        values are iterables containing the column values.

    Returns
    -------
    dict
        Formatted dictionary with the columns of the mmcif file.
    """
    subdict = {k: [_format_string(s) for s in v] for k, v in subdict.items()}
    key_length = {
        key: len(max(value, key=lambda x: len(x), default=""))
        for key, value in subdict.items()
    }
    padded_subdict = {
        key: [s.ljust(key_length[key] + 1) for s in values]
        for key, values in subdict.items()
    }
    return padded_subdict


def create_mask(mask_type: str, **kwargs) -> NDArray:
    """
    Creates a mask of the specified type.

    Parameters
    ----------
    mask_type : str
        Type of the mask to be created. Can be "ellipse", "box", or "tube".
    kwargs : dict
        Additional parameters required by the mask creating functions.

    Returns
    -------
    NDArray
        The created mask.

    Raises
    ------
    ValueError
        If the mask_type is invalid.

    See Also
    --------
    :py:meth:`elliptical_mask`
    :py:meth:`box_mask`
    :py:meth:`tube_mask`
    """
    mapping = {"ellipse": elliptical_mask, "box": box_mask, "tube": tube_mask}
    if mask_type not in mapping:
        raise ValueError(f"mask_type has to be one of {','.join(mapping.keys())}")

    return mapping[mask_type](**kwargs)


def elliptical_mask(
    shape: Tuple[int], radius: Tuple[float], center: Tuple[int]
) -> NDArray:
    """
    Creates an ellipsoidal mask.

    Parameters
    ----------
    shape : tuple
        Shape of the mask to be created.
    radius : tuple
        Radius of the ellipse.
    center : tuple
        Center of the ellipse.

    Returns
    -------
    NDArray
        The created ellipsoidal mask.

    Raises
    ------
    ValueError
        If the length of center and radius is not one or the same as shape.

    Examples
    --------
    >>> mask = elliptical_mask(shape = (20,20), radius = (5,5), center = (10,10))
    """
    center, shape, radius = np.asarray(center), np.asarray(shape), np.asarray(radius)

    radius = np.repeat(radius, shape.size // radius.size)
    center = np.repeat(center, shape.size // center.size)

    if radius.size != shape.size:
        raise ValueError("Length of radius has to be either one or match shape.")
    if center.size != shape.size:
        raise ValueError("Length of center has to be either one or match shape.")

    n = shape.size
    center = center.reshape((-1,) + (1,) * n)
    radius = radius.reshape((-1,) + (1,) * n)

    mask = np.linalg.norm((np.indices(shape) - center) / radius, axis=0)
    mask = (mask <= 1).astype(int)

    return mask


def box_mask(shape: Tuple[int], center: Tuple[int], height: Tuple[int]) -> np.ndarray:
    """
    Creates a box mask centered around the provided center point.

    Parameters
    ----------
    shape : Tuple[int]
        Shape of the output array.
    center : Tuple[int]
        Center point coordinates of the box.
    height : Tuple[int]
        Height (side length) of the box along each axis.

    Returns
    -------
    NDArray
        The created box mask.
    """
    if len(shape) != len(center) or len(center) != len(height):
        raise ValueError("The length of shape, center, and height must be consistent.")

    # Calculate min and max coordinates for the box using the center and half-heights
    center, height = np.array(center, dtype=int), np.array(height, dtype=int)

    half_heights = height // 2
    starts = np.maximum(center - half_heights, 0)
    stops = np.minimum(center + half_heights + np.mod(height, 2) + 1, shape)
    slice_indices = tuple(slice(*coord) for coord in zip(starts, stops))

    out = np.zeros(shape)
    out[slice_indices] = 1
    return out


def tube_mask(
    shape: Tuple[int],
    symmetry_axis: int,
    base_center: Tuple[int],
    inner_radius: float,
    outer_radius: float,
    height: int,
) -> NDArray:
    """
    Creates a tube mask.

    Parameters
    ----------
    shape : tuple
        Shape of the mask to be created.
    symmetry_axis : int
        The axis of symmetry for the tube.
    base_center : tuple
        Center of the base circle of the tube.
    inner_radius : float
        Inner radius of the tube.
    outer_radius : float
        Outer radius of the tube.
    height : int
        Height of the tube.

    Returns
    -------
    NDArray
        The created tube mask.

    Raises
    ------
    ValueError
        If the inner radius is larger than the outer radius. Or height is larger
        than the symmetry axis shape.
    """
    if inner_radius > outer_radius:
        raise ValueError("inner_radius should be smaller than outer_radius.")

    if height > shape[symmetry_axis]:
        raise ValueError(f"Height can be no larger than {shape[symmetry_axis]}.")

    if symmetry_axis > len(shape):
        raise ValueError(f"symmetry_axis can be not larger than {len(shape)}.")

    circle_shape = tuple(b for ix, b in enumerate(shape) if ix != symmetry_axis)
    base_center = tuple(b for ix, b in enumerate(base_center) if ix != symmetry_axis)

    inner_circle = create_mask(
        mask_type="ellipse",
        shape=circle_shape,
        radius=inner_radius,
        center=base_center,
    )
    outer_circle = create_mask(
        mask_type="ellipse",
        shape=circle_shape,
        radius=outer_radius,
        center=base_center,
    )
    circle = outer_circle - inner_circle
    circle = np.expand_dims(circle, axis=symmetry_axis)

    center = shape[symmetry_axis] // 2
    start_idx = center - height // 2
    stop_idx = center + height // 2 + height % 2

    slice_indices = tuple(
        slice(None) if i != symmetry_axis else slice(start_idx, stop_idx)
        for i in range(len(shape))
    )
    tube = np.zeros(shape)
    tube[slice_indices] = np.repeat(circle, height, axis=symmetry_axis)

    return tube


def scramble_phases(
    arr: NDArray, noise_proportion: float = 0.5, seed: int = 42
) -> NDArray:
    """
    Applies random phase scrambling to a given array.

    This function takes an input array, applies a Fourier transform, then scrambles the
    phase with a given proportion of noise, and finally applies an
    inverse Fourier transform to the scrambled data. The phase scrambling
    is controlled by a random seed.

    Parameters
    ----------
    arr : NDArray
        The input array to be scrambled.
    noise_proportion : float, optional
        The proportion of noise in the phase scrambling, by default 0.5.
    seed : int, optional
        The seed for the random phase scrambling, by default 42.

    Returns
    -------
    NDArray
        The array with scrambled phases.

    Raises
    ------
    ValueError
        If noise_proportion is not within [0, 1].
    """
    if noise_proportion < 0 or noise_proportion > 1:
        raise ValueError("noise_proportion has to be within [0, 1].")

    arr_fft = np.fft.fftn(arr)

    amp = np.abs(arr_fft)
    ph = np.angle(arr_fft)

    np.random.seed(seed)
    ph_noise = np.random.permutation(ph)
    ph_new = ph * (1 - noise_proportion) + ph_noise * noise_proportion
    ret = np.real(np.fft.ifftn(amp * np.exp(1j * ph_new)))
    return ret


def conditional_execute(func: Callable, execute_operation: bool = True) -> Callable:
    """
    Return the given function or a no-operation function based on execute_operation.

    Parameters
    ----------
    func : callable
        The function to be executed if execute_operation is True.
    execute_operation : bool, optional
        A flag that determines whether to return `func` or a no-operation function.
        Default is True.

    Returns
    -------
    callable
        Either the given function `func` or a no-operation function.

    Examples
    --------
    >>> def greet(name):
    ...     return f"Hello, {name}!"
    ...
    >>> operation = conditional_execute(greet, False)
    >>> operation("Alice")
    >>> operation = conditional_execute(greet, True)
    >>> operation("Alice")
    'Hello, Alice!'
    """

    def noop(*args, **kwargs):
        """No operation function."""
        pass

    return func if execute_operation else noop



""" Implements template matching based on different metrics. The functions
    implemented here are equivalent to what is implemented in `Fitter`,
    but removing them out of the Fitter classes reduces the time it takes
    when initializing the multiprocessing framework. This is useful in cases,
    where the number of sampled rotations is low, but the number of input
    arrays is high, e.g. split fitting and evaluating different templates.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import os
import sys
from copy import deepcopy
from itertools import product
from typing import Callable, Tuple, Dict
from functools import wraps
from joblib import Parallel, delayed
from multiprocessing.managers import SharedMemoryManager

import numpy as np
from scipy.ndimage import laplace

from .analyzer import MaxScoreOverRotations
from .matching_utils import (
    apply_convolution_mode,
    handle_traceback,
    split_numpy_array_slices,
    conditional_execute,
)
from .matching_memory import MatchingMemoryUsage, MATCHING_MEMORY_REGISTRY
from .preprocessor import Preprocessor
from .matching_data import MatchingData
from .backends import backend
from .types import NDArray, CallbackClass

os.environ["MKL_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["PYFFTW_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"


def _run_inner(backend_name, backend_args, **kwargs):
    from tme.backends import backend

    backend.change_backend(backend_name, **backend_args)
    # print("make sure to remove this")
    # from tme.scoring import scan
    return scan(**kwargs)


def cc_setup(
    rfftn: Callable,
    irfftn: Callable,
    template: NDArray,
    target: NDArray,
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    real_dtype: type,
    complex_dtype: type,
    shared_memory_handler: Callable,
    callback_class: Callable,
    callback_class_args: Dict,
    **kwargs,
) -> Dict:
    """
    Setup to compute the cross-correlation between a target f and template g
    defined as:

    .. math::

        \\mathcal{F}^{-1}(\\mathcal{F}(f) \\cdot \\mathcal{F}(g)^*)


    See Also
    --------
    :py:meth:`corr_scoring`
    :py:class:`tme.matching_optimization.CrossCorrelation`
    """
    target_shape = target.shape
    target_pad = backend.topleft_pad(target, fast_shape)
    target_pad_ft = backend.preallocate_array(fast_ft_shape, complex_dtype)

    rfftn(target_pad, target_pad_ft)

    target_ft_out = backend.arr_to_sharedarr(
        arr=target_pad_ft, shared_memory_handler=shared_memory_handler
    )

    template_out = backend.arr_to_sharedarr(
        arr=template, shared_memory_handler=shared_memory_handler
    )
    inv_denominator_buffer = backend.arr_to_sharedarr(
        arr=backend.preallocate_array(1, real_dtype) + 1,
        shared_memory_handler=shared_memory_handler,
    )
    numerator2_buffer = backend.arr_to_sharedarr(
        arr=backend.preallocate_array(1, real_dtype),
        shared_memory_handler=shared_memory_handler,
    )

    target_ft_tuple = (target_ft_out, fast_ft_shape, complex_dtype)
    template_tuple = (template_out, template.shape, real_dtype)

    inv_denominator_tuple = (inv_denominator_buffer, (1,), real_dtype)
    numerator2_tuple = (numerator2_buffer, (1,), real_dtype)

    ret = {
        "template": template_tuple,
        "ft_target": target_ft_tuple,
        "inv_denominator": inv_denominator_tuple,
        "numerator2": numerator2_tuple,
        "targetshape": target_shape,
        "templateshape": template.shape,
        "fast_shape": fast_shape,
        "fast_ft_shape": fast_ft_shape,
        "real_dtype": real_dtype,
        "complex_dtype": complex_dtype,
        "callback_class": callback_class,
        "callback_class_args": callback_class_args,
        "use_memmap": kwargs.get("use_memmap", False),
        "temp_dir": kwargs.get("temp_dir", None),
    }

    return ret


def lcc_setup(**kwargs) -> Dict:
    """
    Setup to compute the cross-correlation between a laplace transformed target f
    and laplace transformed template g defined as:

    .. math::

        \\mathcal{F}^{-1}(\\mathcal{F}(\\nabla^{2}f) \\cdot \\mathcal{F}(\\nabla^{2} g)^*)


    See Also
    --------
    :py:meth:`corr_scoring`
    :py:class:`tme.matching_optimization.LaplaceCrossCorrelation`
    """
    from cupyx.scipy.ndimage import laplace
    kwargs["target"] = laplace(kwargs["target"], mode="wrap")
    kwargs["template"] = laplace(kwargs["template"], mode="wrap")
    return flcSphericalMask_setup(**kwargs)


def corr_setup(
    rfftn: Callable,
    irfftn: Callable,
    template: NDArray,
    template_mask: NDArray,
    target: NDArray,
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    real_dtype: type,
    complex_dtype: type,
    shared_memory_handler: Callable,
    callback_class: Callable,
    callback_class_args: Dict,
    **kwargs,
) -> Dict:
    """
    Setup to compute a normalized cross-correlation between a target f and a template g.

    .. math::

        \\frac{CC(f,g) - \\overline{g} \\cdot CC(f, m)}
        {(CC(f^2, m) - \\frac{CC(f, m)^2}{N_g}) \\cdot \\sigma_{g}}

    Where:

    .. math::

        CC(f,g) = \\mathcal{F}^{-1}(\\mathcal{F}(f) \\cdot \\mathcal{F}(g)^*)

    and m is a mask with the same dimension as the template filled with ones.

    References
    ----------
    .. [1]  J. P. Lewis, "Fast Normalized Cross-Correlation", Industrial Light
            and Magic.

    See Also
    --------
    :py:meth:`corr_scoring`
    :py:class:`tme.matching_optimization.NormalizedCrossCorrelation`.
    """
    target_pad = backend.topleft_pad(target, fast_shape)

    # The exact composition of the denominator is debatable
    # scikit-image match_template multiplies the running sum of the target
    # with a scaling factor derived from the template. This is probably appropriate
    # in pattern matching situations where the template exists in the target
    window_template = backend.topleft_pad(template_mask, fast_shape)
    ft_window_template = backend.preallocate_array(fast_ft_shape, complex_dtype)
    rfftn(window_template, ft_window_template)
    window_template = None

    # Target and squared target window sums
    ft_target = backend.preallocate_array(fast_ft_shape, complex_dtype)
    ft_target2 = backend.preallocate_array(fast_ft_shape, complex_dtype)
    denominator = backend.preallocate_array(fast_shape, real_dtype)
    target_window_sum = backend.preallocate_array(fast_shape, real_dtype)
    rfftn(target_pad, ft_target)

    rfftn(backend.square(target_pad), ft_target2)
    backend.multiply(ft_target2, ft_window_template, out=ft_target2)
    irfftn(ft_target2, denominator)

    backend.multiply(ft_target, ft_window_template, out=ft_window_template)
    irfftn(ft_window_template, target_window_sum)

    target_pad, ft_target2, ft_window_template = None, None, None

    # Normalizing constants
    template_mean = backend.mean(template)
    template_volume = np.prod(template.shape)
    template_ssd = backend.sum(
        backend.square(backend.subtract(template, template_mean))
    )

    # Final numerator is score - numerator2
    numerator2 = backend.multiply(target_window_sum, template_mean)

    # Compute denominator
    backend.multiply(target_window_sum, target_window_sum, out=target_window_sum)
    backend.divide(target_window_sum, template_volume, out=target_window_sum)

    backend.subtract(denominator, target_window_sum, out=denominator)
    backend.multiply(denominator, template_ssd, out=denominator)
    backend.maximum(denominator, 0, out=denominator)
    backend.sqrt(denominator, out=denominator)
    target_window_sum = None

    # Invert denominator to compute final score as product
    denominator_mask = denominator > backend.eps(denominator.dtype)
    inv_denominator = backend.preallocate_array(fast_shape, real_dtype)
    inv_denominator[denominator_mask] = 1 / denominator[denominator_mask]

    # Convert arrays used in subsequent fitting to SharedMemory objects
    template_buffer = backend.arr_to_sharedarr(
        arr=template, shared_memory_handler=shared_memory_handler
    )
    target_ft_buffer = backend.arr_to_sharedarr(
        arr=ft_target, shared_memory_handler=shared_memory_handler
    )
    inv_denominator_buffer = backend.arr_to_sharedarr(
        arr=inv_denominator, shared_memory_handler=shared_memory_handler
    )
    numerator2_buffer = backend.arr_to_sharedarr(
        arr=numerator2, shared_memory_handler=shared_memory_handler
    )

    template_tuple = (template_buffer, deepcopy(template.shape), real_dtype)
    target_ft_tuple = (target_ft_buffer, fast_ft_shape, complex_dtype)

    inv_denominator_tuple = (inv_denominator_buffer, fast_shape, real_dtype)
    numerator2_tuple = (numerator2_buffer, fast_shape, real_dtype)

    ft_target, inv_denominator, numerator2 = None, None, None

    ret = {
        "template": template_tuple,
        "ft_target": target_ft_tuple,
        "inv_denominator": inv_denominator_tuple,
        "numerator2": numerator2_tuple,
        "targetshape": deepcopy(target.shape),
        "templateshape": deepcopy(template.shape),
        "fast_shape": fast_shape,
        "fast_ft_shape": fast_ft_shape,
        "real_dtype": real_dtype,
        "complex_dtype": complex_dtype,
        "callback_class": callback_class,
        "callback_class_args": callback_class_args,
        "template_mean": kwargs.get("template_mean", template_mean),
    }

    return ret


def cam_setup(**kwargs):
    """
    Setup to compute a normalized cross-correlation between a target f and a template g
    over their means. In practice this can be expressed like the cross-correlation
    CORR defined in :py:meth:`corr_scoring`, so that:

    Notes
    -----

    .. math::

        \\text{CORR}(f-\\overline{f}, g - \\overline{g})

    Where

    .. math::

        \\overline{f}, \\overline{g}

    are the mean of f and g respectively.

    References
    ----------
    .. [1]  J. P. Lewis, "Fast Normalized Cross-Correlation", Industrial Light
            and Magic.

    See Also
    --------
    :py:meth:`corr_scoring`.
    :py:class:`tme.matching_optimization.NormalizedCrossCorrelationMean`.
    """
    kwargs["template"] = kwargs["template"] - kwargs["template"].mean()
    return corr_setup(**kwargs)


def flc_setup(
    rfftn: Callable,
    irfftn: Callable,
    template: NDArray,
    template_mask: NDArray,
    target: NDArray,
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    real_dtype: type,
    complex_dtype: type,
    shared_memory_handler: Callable,
    callback_class: Callable,
    callback_class_args: Dict,
    **kwargs,
) -> Dict:
    """
    Setup to compute a normalized cross-correlation score of a target f a template g
    and a mask m:

    .. math::

        \\frac{CC(f, \\frac{g*m - \\overline{g*m}}{\\sigma_{g*m}})}
        {N_m * \\sqrt{
            \\frac{CC(f^2, m)}{N_m} - (\\frac{CC(f, m)}{N_m})^2}
        }

    Where:

    .. math::

        CC(f,g) = \\mathcal{F}^{-1}(\\mathcal{F}(f) \\cdot \\mathcal{F}(g)^*)

    and Nm is the number of voxels within the template mask m.

    References
    ----------
    .. [1]  W. Wan, S. Khavnekar, J. Wagner, P. Erdmann, and W. Baumeister
            Microsc. Microanal. 26, 2516 (2020)
    .. [2]  T. Hrabe, Y. Chen, S. Pfeffer, L. Kuhn Cuellar, A.-V. Mangold,
            and F. Förster, J. Struct. Biol. 178, 177 (2012).

    See Also
    --------
    :py:meth:`flc_scoring`
    """
    target_pad = backend.topleft_pad(target, fast_shape)

    # Target and squared target window sums
    ft_target = backend.preallocate_array(fast_ft_shape, complex_dtype)
    ft_target2 = backend.preallocate_array(fast_ft_shape, complex_dtype)
    rfftn(target_pad, ft_target)
    rfftn(backend.square(target_pad), ft_target2)

    # Convert arrays used in subsequent fitting to SharedMemory objects
    ft_target = backend.arr_to_sharedarr(
        arr=ft_target, shared_memory_handler=shared_memory_handler
    )
    ft_target2 = backend.arr_to_sharedarr(
        arr=ft_target2, shared_memory_handler=shared_memory_handler
    )

    template_mask = template_mask > 0
    template_mean = backend.mean(template[template_mask])
    template_std = backend.std(template[template_mask])
    template_mask = backend.astype(template_mask, real_dtype)

    backend.divide(template - template_mean, template_std, out=template)
    backend.multiply(template, template_mask, out=template)

    template_buffer = backend.arr_to_sharedarr(
        arr=template, shared_memory_handler=shared_memory_handler
    )
    template_mask_buffer = backend.arr_to_sharedarr(
        arr=template_mask, shared_memory_handler=shared_memory_handler
    )

    template_tuple = (template_buffer, template.shape, real_dtype)
    template_mask_tuple = (template_mask_buffer, template_mask.shape, real_dtype)

    target_ft_tuple = (ft_target, fast_ft_shape, complex_dtype)
    target_ft2_tuple = (ft_target2, fast_ft_shape, complex_dtype)

    ret = {
        "template": template_tuple,
        "template_mask": template_mask_tuple,
        "ft_target": target_ft_tuple,
        "ft_target2": target_ft2_tuple,
        "targetshape": target.shape,
        "templateshape": template.shape,
        "fast_shape": fast_shape,
        "fast_ft_shape": fast_ft_shape,
        "real_dtype": real_dtype,
        "complex_dtype": complex_dtype,
        "callback_class": callback_class,
        "callback_class_args": callback_class_args,
    }

    return ret


def flcSphericalMask_setup(
    rfftn: Callable,
    irfftn: Callable,
    template: NDArray,
    template_mask: NDArray,
    target: NDArray,
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    real_dtype: type,
    complex_dtype: type,
    shared_memory_handler: Callable,
    callback_class: Callable,
    callback_class_args: Dict,
    **kwargs,
) -> Dict:
    """
    Like :py:meth:`flc_setup` but for rotation invariant masks. In such cases
    the score can be computed quicker by not computing the fourier transforms
    of the mask for each rotation.

    .. math::

        \\frac{CC(f, \\frac{g*m - \\overline{g*m}}{\\sigma_{g*m}})}
        {N_m * \\sqrt{
            \\frac{CC(f^2, m)}{N_m} - (\\frac{CC(f, m)}{N_m})^2}
        }

    Where:

    .. math::

        CC(f,g) = \\mathcal{F}^{-1}(\\mathcal{F}(f) \\cdot \\mathcal{F}(g)^*)

    and Nm is the number of voxels within the template mask m.

    References
    ----------
    .. [1]  W. Wan, S. Khavnekar, J. Wagner, P. Erdmann, and W. Baumeister
            Microsc. Microanal. 26, 2516 (2020)
    .. [2]  T. Hrabe, Y. Chen, S. Pfeffer, L. Kuhn Cuellar, A.-V. Mangold,
            and F. Förster, J. Struct. Biol. 178, 177 (2012).

    See Also
    --------
    :py:meth:`corr_scoring`
    """
    target_pad = backend.topleft_pad(target, fast_shape)
    template_mask_pad = backend.topleft_pad(template_mask, fast_shape)

    # Target and squared target window sums
    ft_target = backend.preallocate_array(fast_ft_shape, complex_dtype)
    ft_template_mask = backend.preallocate_array(fast_ft_shape, complex_dtype)
    ft_temp = backend.preallocate_array(fast_ft_shape, complex_dtype)

    temp = backend.preallocate_array(fast_shape, real_dtype)
    temp2 = backend.preallocate_array(fast_shape, real_dtype)
    numerator2 = backend.preallocate_array(1, real_dtype)

    eps = backend.eps(real_dtype)
    n_observations = backend.sum(template_mask_pad)
    rfftn(template_mask_pad, ft_template_mask)

    # Variance part denominator
    rfftn(backend.square(target_pad), ft_target)
    backend.multiply(ft_target, ft_template_mask, out=ft_temp)
    irfftn(ft_temp, temp2)
    backend.divide(temp2, n_observations, out=temp2)

    # Mean part denominator
    rfftn(target_pad, ft_target)
    backend.multiply(ft_target, ft_template_mask, out=ft_temp)
    irfftn(ft_temp, temp)
    backend.divide(temp, n_observations, out=temp)
    backend.square(temp, out=temp)

    backend.subtract(temp2, temp, out=temp)

    backend.maximum(temp, 0.0, out=temp)
    backend.sqrt(temp, out=temp)
    backend.multiply(temp, n_observations, out=temp)

    tol = 1e3 * eps * backend.max(backend.abs(temp))
    nonzero_indices = temp > tol

    backend.fill(temp2, 0)
    temp2[nonzero_indices] = 1 / temp[nonzero_indices]

    template_mask = template_mask > 0
    template_mean = backend.mean(template[template_mask])
    template_std = backend.std(template[template_mask])

    template = backend.divide(backend.subtract(template, template_mean), template_std)
    backend.multiply(template, template_mask, out=template)

    # Convert arrays used in subsequent fitting to SharedMemory objects
    template_buffer = backend.arr_to_sharedarr(
        arr=template, shared_memory_handler=shared_memory_handler
    )
    target_ft_buffer = backend.arr_to_sharedarr(
        arr=ft_target, shared_memory_handler=shared_memory_handler
    )
    inv_denominator_buffer = backend.arr_to_sharedarr(
        arr=temp2, shared_memory_handler=shared_memory_handler
    )
    numerator2_buffer = backend.arr_to_sharedarr(
        arr=numerator2, shared_memory_handler=shared_memory_handler
    )

    template_tuple = (template_buffer, template.shape, real_dtype)
    target_ft_tuple = (target_ft_buffer, fast_ft_shape, complex_dtype)

    inv_denominator_tuple = (inv_denominator_buffer, fast_shape, real_dtype)
    numerator2_tuple = (numerator2_buffer, (1,), real_dtype)

    ret = {
        "template": template_tuple,
        "ft_target": target_ft_tuple,
        "inv_denominator": inv_denominator_tuple,
        "numerator2": numerator2_tuple,
        "targetshape": target.shape,
        "templateshape": template.shape,
        "fast_shape": fast_shape,
        "fast_ft_shape": fast_ft_shape,
        "real_dtype": real_dtype,
        "complex_dtype": complex_dtype,
        "callback_class": callback_class,
        "callback_class_args": callback_class_args,
    }

    return ret


def mcc_setup(
    rfftn: Callable,
    irfftn: Callable,
    template: NDArray,
    template_mask: NDArray,
    target: NDArray,
    target_mask: NDArray,
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    real_dtype: type,
    complex_dtype: type,
    shared_memory_handler: Callable,
    callback_class: Callable,
    callback_class_args: Dict,
    **kwargs,
) -> Dict:
    """
    Setup to compute a normalized cross-correlation score with masks
    for both template and target.

    .. math::

        \\frac{
               CC(f, g) - \\frac{CC(f, m) \\cdot CC(t, g)}{CC(t, m)}
            }{
            \\sqrt{
                (CC(f ^ 2, m) - \\frac{CC(f, m) ^ 2}{CC(t, m)}) \\cdot
                (CC(t, g^2) - \\frac{CC(t, g) ^ 2}{CC(t, m)})
                }
            }

    Where:

    .. math::

        CC(f,g) = \\mathcal{F}^{-1}(\\mathcal{F}(f) \\cdot \\mathcal{F}(g)^*)


    References
    ----------
    .. [1]  Masked FFT registration, Dirk Padfield, CVPR 2010 conference

    See Also
    --------
    :py:meth:`mcc_scoring`
    :py:class:`tme.matching_optimization.MaskedCrossCorrelation`
    """
    target = backend.multiply(target, target_mask > 0, out=target)

    target_pad = backend.topleft_pad(target, fast_shape)
    target_mask_pad = backend.topleft_pad(target_mask, fast_shape)

    target_ft = backend.preallocate_array(fast_ft_shape, complex_dtype)
    rfftn(target_pad, target_ft)
    target_ft_buffer = backend.arr_to_sharedarr(
        arr=target_ft, shared_memory_handler=shared_memory_handler
    )

    target_ft2 = backend.preallocate_array(fast_ft_shape, complex_dtype)
    rfftn(backend.square(target_pad), target_ft2)
    target_ft2_buffer = backend.arr_to_sharedarr(
        arr=target_ft2, shared_memory_handler=shared_memory_handler
    )

    target_mask_ft = backend.preallocate_array(fast_ft_shape, complex_dtype)
    rfftn(target_mask_pad, target_mask_ft)
    target_mask_ft_buffer = backend.arr_to_sharedarr(
        arr=target_mask_ft, shared_memory_handler=shared_memory_handler
    )

    template_buffer = backend.arr_to_sharedarr(
        arr=template, shared_memory_handler=shared_memory_handler
    )
    template_mask_buffer = backend.arr_to_sharedarr(
        arr=template_mask, shared_memory_handler=shared_memory_handler
    )

    template_tuple = (template_buffer, template.shape, real_dtype)
    template_mask_tuple = (template_mask_buffer, template.shape, real_dtype)

    target_ft_tuple = (target_ft_buffer, fast_ft_shape, complex_dtype)
    target_ft2_tuple = (target_ft2_buffer, fast_ft_shape, complex_dtype)
    target_mask_ft_tuple = (target_mask_ft_buffer, fast_ft_shape, complex_dtype)

    ret = {
        "template": template_tuple,
        "template_mask": template_mask_tuple,
        "ft_target": target_ft_tuple,
        "ft_target2": target_ft2_tuple,
        "ft_target_mask": target_mask_ft_tuple,
        "targetshape": target.shape,
        "templateshape": template.shape,
        "fast_shape": fast_shape,
        "fast_ft_shape": fast_ft_shape,
        "real_dtype": real_dtype,
        "complex_dtype": complex_dtype,
        "callback_class": callback_class,
        "callback_class_args": callback_class_args,
    }

    return ret


def corr_scoring(
    template: Tuple[type, Tuple[int], type],
    ft_target: Tuple[type, Tuple[int], type],
    inv_denominator: Tuple[type, Tuple[int], type],
    numerator2: Tuple[type, Tuple[int], type],
    template_filter: Tuple[type, Tuple[int], type],
    targetshape: Tuple[int],
    templateshape: Tuple[int],
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    rotations: NDArray,
    real_dtype: type,
    complex_dtype: type,
    callback_class: CallbackClass,
    callback_class_args: Dict,
    interpolation_order: int,
    convolution_mode: str = "full",
    **kwargs,
) -> CallbackClass:
    """
    Calculates a normalized cross-correlation between a target f and a template g.

    .. math::

        (CC(f,g) - numerator2) \\cdot inv\\_denominator

    Parameters
    ----------
    template : Tuple[type, Tuple[int], type]
        Tuple containing a pointer to the template data, its shape, and its datatype.
    ft_target : Tuple[type, Tuple[int], type]
        Tuple containing a pointer to the fourier tranform of the target,
        its shape, and its datatype.
    inv_denominator : Tuple[type, Tuple[int], type]
        Tuple containing a pointer to the inverse denominator data, its shape, and its
        datatype.
    numerator2 : Tuple[type, Tuple[int], type]
        Tuple containing a pointer to the numerator2 data, its shape, and its datatype.
    targetshape : Tuple[int]
        The shape of the target.
    templateshape : Tuple[int]
        The shape of the template.
    fast_shape : Tuple[int]
        The shape for fast Fourier transform.
    fast_ft_shape : Tuple[int]
        The shape for fast Fourier transform of the target.
    rotations : NDArray
        Array containing the rotation matrices to be applied on the template.
    real_dtype : type
        Data type for the real part of the array.
    complex_dtype : type
        Data type for the complex part of the array.
    callback_class : CallbackClass
        A callable class or function for processing the results after each
        rotation.
    callback_class_args : Dict
        Dictionary of arguments to be passed to the callback class if it's
        instantiable.
    interpolation_order : int
        The order of interpolation to be used while rotating the template.
    convolution_mode : str, optional
        Mode to use for convolution, default is "full".
    **kwargs :
        Additional arguments to be passed to the function.

    Returns
    -------
    CallbackClass
        If callback_class was provided an instance of callback_class otherwise None.

    See Also
    --------
    :py:meth:`cc_setup`
    :py:meth:`corr_setup`
    :py:meth:`cam_setup`
    :py:meth:`flcSphericalMask_setup`
    """
    template_buffer, template_shape, template_dtype = template
    ft_target_buffer, ft_target_shape, ft_target_dtype = ft_target
    inv_denominator_buffer, inv_denominator_pointer_shape, _ = inv_denominator
    numerator2_buffer, numerator2_shape, _ = numerator2
    filter_buffer, filter_shape, filter_dtype = template_filter

    if callback_class is not None and isinstance(callback_class, type):
        callback = callback_class(**callback_class_args)
    elif not isinstance(callback_class, type):
        callback = callback_class

    # Retrieve objects from shared memory
    template = backend.sharedarr_to_arr(template_shape, template_dtype, template_buffer)
    ft_target = backend.sharedarr_to_arr(
        ft_target_shape, ft_target_dtype, ft_target_buffer
    )
    inv_denominator = backend.sharedarr_to_arr(
        inv_denominator_pointer_shape, template_dtype, inv_denominator_buffer
    )
    numerator2 = backend.sharedarr_to_arr(
        numerator2_shape, template_dtype, numerator2_buffer
    )
    template_filter = backend.sharedarr_to_arr(
        filter_shape, filter_dtype, filter_buffer
    )

    arr = backend.preallocate_array(fast_shape, real_dtype)
    ft_temp = backend.preallocate_array(fast_ft_shape, complex_dtype)

    rfftn, irfftn = backend.build_fft(
        fast_shape=fast_shape,
        fast_ft_shape=fast_ft_shape,
        real_dtype=real_dtype,
        complex_dtype=complex_dtype,
        fftargs=kwargs.get("fftargs", {}),
        temp_real=arr,
        temp_fft=ft_temp,
    )

    norm_numerator = (backend.sum(numerator2) != 0) & (backend.size(numerator2) != 1)
    norm_denominator = (backend.sum(inv_denominator) != 1) & (
        backend.size(inv_denominator) != 1
    )
    filter_template = backend.size(template_filter) != 0

    norm_func_numerator = conditional_execute(backend.subtract, norm_numerator)
    norm_func_denominator = conditional_execute(backend.multiply, norm_denominator)
    template_filter_func = conditional_execute(backend.multiply, filter_template)

    axis = tuple(range(arr.ndim))
    fourier_shift = callback_class_args.get("fourier_shift", backend.zeros(arr.ndim))
    fourier_shift_scores = backend.sum(fourier_shift != 0) != 0

    for index in range(rotations.shape[0]):
        rotation = rotations[index]
        backend.fill(arr, 0)
        backend.rotate_array(
            arr=template,
            rotation_matrix=rotation,
            out=arr,
            use_geometric_center=False,
            order=interpolation_order,
        )
        rfftn(arr, ft_temp)
        template_filter_func(ft_temp, template_filter, out=ft_temp)

        backend.multiply(ft_target, ft_temp, out=ft_temp)
        irfftn(ft_temp, arr)

        norm_func_numerator(arr, numerator2, out=arr)
        norm_func_denominator(arr, inv_denominator, out=arr)

        if fourier_shift_scores:
            arr = backend.roll(arr, shift=fourier_shift, axis=axis)

        score = apply_convolution_mode(
            arr, convolution_mode=convolution_mode, s1=targetshape, s2=templateshape
        )

        if callback_class is not None:
            callback(
                score,
                rotation_matrix=rotation,
                rotation_index=index,
                **callback_class_args,
            )

    return callback


def flc_scoring(
    template: Tuple[type, Tuple[int], type],
    template_mask: Tuple[type, Tuple[int], type],
    ft_target: Tuple[type, Tuple[int], type],
    ft_target2: Tuple[type, Tuple[int], type],
    template_filter: Tuple[type, Tuple[int], type],
    targetshape: Tuple[int],
    templateshape: Tuple[int],
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    rotations: NDArray,
    real_dtype: type,
    complex_dtype: type,
    callback_class: CallbackClass,
    callback_class_args: Dict,
    interpolation_order: int,
    **kwargs,
) -> CallbackClass:
    """
    Computes a normalized cross-correlation score of a target f a template g
    and a mask m:

    .. math::

        \\frac{CC(f, \\frac{g*m - \\overline{g*m}}{\\sigma_{g*m}})}
        {N_m * \\sqrt{
            \\frac{CC(f^2, m)}{N_m} - (\\frac{CC(f, m)}{N_m})^2}
        }

    Where:

    .. math::

        CC(f,g) = \\mathcal{F}^{-1}(\\mathcal{F}(f) \\cdot \\mathcal{F}(g)^*)

    and Nm is the number of voxels within the template mask m.

    References
    ----------
    .. [1]  W. Wan, S. Khavnekar, J. Wagner, P. Erdmann, and W. Baumeister
            Microsc. Microanal. 26, 2516 (2020)
    .. [2]  T. Hrabe, Y. Chen, S. Pfeffer, L. Kuhn Cuellar, A.-V. Mangold,
            and F. Förster, J. Struct. Biol. 178, 177 (2012).
    """
    template_buffer, template_shape, template_dtype = template
    template_mask_buffer, *_ = template_mask
    filter_buffer, filter_shape, filter_dtype = template_filter

    ft_target_buffer, ft_target_shape, ft_target_dtype = ft_target
    ft_target2_buffer, *_ = ft_target2

    if callback_class is not None and isinstance(callback_class, type):
        callback = callback_class(**callback_class_args)
    elif not isinstance(callback_class, type):
        callback = callback_class

    # Retrieve objects from shared memory
    template = backend.sharedarr_to_arr(template_shape, template_dtype, template_buffer)
    template_mask = backend.sharedarr_to_arr(
        template_shape, template_dtype, template_mask_buffer
    )
    ft_target = backend.sharedarr_to_arr(
        ft_target_shape, ft_target_dtype, ft_target_buffer
    )
    ft_target2 = backend.sharedarr_to_arr(
        ft_target_shape, ft_target_dtype, ft_target2_buffer
    )
    template_filter = backend.sharedarr_to_arr(
        filter_shape, filter_dtype, filter_buffer
    )

    arr = backend.preallocate_array(fast_shape, real_dtype)
    temp = backend.preallocate_array(fast_shape, real_dtype)
    temp2 = backend.preallocate_array(fast_shape, real_dtype)

    ft_temp = backend.preallocate_array(fast_ft_shape, complex_dtype)
    ft_denom = backend.preallocate_array(fast_ft_shape, complex_dtype)

    rfftn, irfftn = backend.build_fft(
        fast_shape=fast_shape,
        fast_ft_shape=fast_ft_shape,
        real_dtype=real_dtype,
        complex_dtype=complex_dtype,
        fftargs=kwargs.get("fftargs", {}),
        temp_real=arr,
        temp_fft=ft_temp,
    )
    eps = backend.eps(real_dtype)
    filter_template = backend.size(template_filter) != 0
    template_filter_func = conditional_execute(backend.multiply, filter_template)

    axis = tuple(range(arr.ndim))
    fourier_shift = callback_class_args.get("fourier_shift", backend.zeros(arr.ndim))
    fourier_shift_scores = backend.sum(fourier_shift != 0) != 0

    for index in range(rotations.shape[0]):
        rotation = rotations[index]
        backend.fill(arr, 0)
        backend.fill(temp, 0)
        backend.rotate_array(
            arr=template,
            arr_mask=template_mask,
            rotation_matrix=rotation,
            out=arr,
            out_mask=temp,
            use_geometric_center=False,
            order=interpolation_order,
        )
        n_observations = backend.sum(temp)

        rfftn(temp, ft_temp)

        backend.multiply(ft_target, ft_temp, out=ft_denom)
        irfftn(ft_denom, temp)
        backend.divide(temp, n_observations, out=temp)
        backend.square(temp, out=temp)

        backend.multiply(ft_target2, ft_temp, out=ft_denom)
        irfftn(ft_denom, temp2)
        backend.divide(temp2, n_observations, out=temp2)

        backend.subtract(temp2, temp, out=temp)
        backend.maximum(temp, 0.0, out=temp)
        backend.sqrt(temp, out=temp)
        backend.multiply(temp, n_observations, out=temp)

        rfftn(arr, ft_temp)
        template_filter_func(ft_temp, template_filter, out=ft_temp)
        backend.multiply(ft_target, ft_temp, out=ft_temp)
        irfftn(ft_temp, arr)

        tol = tol = 1e3 * eps * backend.max(backend.abs(temp))
        nonzero_indices = temp > tol
        backend.fill(temp2, 0)
        temp2[nonzero_indices] = arr[nonzero_indices] / temp[nonzero_indices]

        convolution_mode = kwargs.get("convolution_mode", "full")

        if fourier_shift_scores:
            temp2 = backend.roll(temp2, shift=fourier_shift, axis=axis)

        score = apply_convolution_mode(
            temp2, convolution_mode=convolution_mode, s1=targetshape, s2=templateshape
        )

        if callback_class is not None:
            callback(
                score,
                rotation_matrix=rotation,
                rotation_index=index,
                **callback_class_args,
            )

    return callback


def mcc_scoring(
    template: Tuple[type, Tuple[int], type],
    template_mask: Tuple[type, Tuple[int], type],
    ft_target: Tuple[type, Tuple[int], type],
    ft_target2: Tuple[type, Tuple[int], type],
    ft_target_mask: Tuple[type, Tuple[int], type],
    template_filter: Tuple[type, Tuple[int], type],
    targetshape: Tuple[int],
    templateshape: Tuple[int],
    fast_shape: Tuple[int],
    fast_ft_shape: Tuple[int],
    rotations: NDArray,
    real_dtype: type,
    complex_dtype: type,
    callback_class: CallbackClass,
    callback_class_args: type,
    interpolation_order: int,
    overlap_ratio: float = 0.3,
    **kwargs,
) -> CallbackClass:
    """
    Computes a cross-correlation score with masks for both template and target.

    .. math::

        \\frac{
               CC(f, g) - \\frac{CC(f, m) \\cdot CC(t, g)}{CC(t, m)}
            }{
            \\sqrt{
                (CC(f ^ 2, m) - \\frac{CC(f, m) ^ 2}{CC(t, m)}) \\cdot
                (CC(t, g^2) - \\frac{CC(t, g) ^ 2}{CC(t, m)})
                }
            }

    Where:

    .. math::

        CC(f,g) = \\mathcal{F}^{-1}(\\mathcal{F}(f) \\cdot \\mathcal{F}(g)^*)


    References
    ----------
    .. [1]  Masked FFT registration, Dirk Padfield, CVPR 2010 conference
    .. [2]  https://scikit-image.org/docs/stable/api/skimage.registration.html

    See Also
    --------
    :py:class:`tme.matching_optimization.MaskedCrossCorrelation`
    """
    template_buffer, template_shape, template_dtype = template
    ft_target_buffer, ft_target_shape, ft_target_dtype = ft_target
    ft_target2_buffer, ft_target_shape, ft_target_dtype = ft_target2
    template_mask_buffer, _, _ = template
    ft_target_mask_buffer, _, _ = ft_target
    filter_buffer, filter_shape, filter_dtype = template_filter

    if callback_class is not None and isinstance(callback_class, type):
        callback = callback_class(**callback_class_args)
    elif not isinstance(callback_class, type):
        callback = callback_class

    # Retrieve objects from shared memory
    template = backend.sharedarr_to_arr(template_shape, template_dtype, template_buffer)
    target_ft = backend.sharedarr_to_arr(
        ft_target_shape, ft_target_dtype, ft_target_buffer
    )
    target_ft2 = backend.sharedarr_to_arr(
        ft_target_shape, ft_target_dtype, ft_target2_buffer
    )
    template_mask = backend.sharedarr_to_arr(
        template_shape, template_dtype, template_mask_buffer
    )
    target_mask_ft = backend.sharedarr_to_arr(
        ft_target_shape, ft_target_dtype, ft_target_mask_buffer
    )
    template_filter = backend.sharedarr_to_arr(
        filter_shape, filter_dtype, filter_buffer
    )

    axes = tuple(range(template.ndim))
    eps = backend.eps(real_dtype)

    # Allocate score and process specific arrays
    template_rot = backend.preallocate_array(fast_shape, real_dtype)
    mask_overlap = backend.preallocate_array(fast_shape, real_dtype)
    numerator = backend.preallocate_array(fast_shape, real_dtype)

    temp = backend.preallocate_array(fast_shape, real_dtype)
    temp2 = backend.preallocate_array(fast_shape, real_dtype)
    temp3 = backend.preallocate_array(fast_shape, real_dtype)
    temp_ft = backend.preallocate_array(fast_ft_shape, complex_dtype)

    rfftn, irfftn = backend.build_fft(
        fast_shape=fast_shape,
        fast_ft_shape=fast_ft_shape,
        real_dtype=real_dtype,
        complex_dtype=complex_dtype,
        fftargs=kwargs.get("fftargs", {}),
        temp_real=numerator,
        temp_fft=temp_ft,
    )

    filter_template = backend.size(template_filter) != 0
    template_filter_func = conditional_execute(backend.multiply, filter_template)

    axis = tuple(range(template.ndim))
    fourier_shift = callback_class_args.get(
        "fourier_shift", backend.zeros(template.ndim)
    )
    fourier_shift_scores = backend.sum(fourier_shift != 0) != 0

    # Calculate scores across all rotations
    for index in range(rotations.shape[0]):
        rotation = rotations[index]
        backend.fill(template_rot, 0)
        backend.fill(temp, 0)

        backend.rotate_array(
            arr=template,
            arr_mask=template_mask,
            rotation_matrix=rotation,
            out=template_rot,
            out_mask=temp,
            use_geometric_center=False,
            order=interpolation_order,
        )

        backend.multiply(template_rot, temp > 0, out=template_rot)

        # template_rot_ft
        rfftn(template_rot, temp_ft)
        template_filter_func(temp_ft, template_filter, out=temp_ft)
        irfftn(target_mask_ft * temp_ft, temp2)
        irfftn(target_ft * temp_ft, numerator)

        # temp template_mask_rot | temp_ft template_mask_rot_ft
        # Calculate overlap of masks at every point in the convolution.
        # Locations with high overlap should not be taken into account.
        rfftn(temp, temp_ft)
        irfftn(temp_ft * target_mask_ft, mask_overlap)
        mask_overlap[:] = np.round(mask_overlap)
        mask_overlap[:] = np.maximum(mask_overlap, eps)
        irfftn(temp_ft * target_ft, temp)

        backend.subtract(
            numerator,
            backend.divide(backend.multiply(temp, temp2), mask_overlap),
            out=numerator,
        )

        # temp_3 = fixed_denom
        backend.multiply(temp_ft, target_ft2, out=temp_ft)
        irfftn(temp_ft, temp3)
        backend.subtract(
            temp3, backend.divide(backend.square(temp), mask_overlap), out=temp3
        )
        backend.maximum(temp3, 0.0, out=temp3)

        # temp = moving_denom
        rfftn(backend.square(template_rot), temp_ft)
        backend.multiply(target_mask_ft, temp_ft, out=temp_ft)
        irfftn(temp_ft, temp)

        backend.subtract(
            temp, backend.divide(backend.square(temp2), mask_overlap), out=temp
        )
        backend.maximum(temp, 0.0, out=temp)

        # temp_2 = denom
        backend.multiply(temp3, temp, out=temp)
        backend.sqrt(temp, temp2)

        # Pixels where `denom` is very small will introduce large
        # numbers after division. To get around this problem,
        # we zero-out problematic pixels.
        tol = 1e3 * eps * backend.max(backend.abs(temp2), axis=axes, keepdims=True)
        nonzero_indices = temp2 > tol

        backend.fill(temp, 0)
        temp[nonzero_indices] = numerator[nonzero_indices] / temp2[nonzero_indices]
        backend.clip(temp, a_min=-1, a_max=1, out=temp)

        # Apply overlap ratio threshold
        number_px_threshold = overlap_ratio * backend.max(
            mask_overlap, axis=axes, keepdims=True
        )
        temp[mask_overlap < number_px_threshold] = 0.0
        convolution_mode = kwargs.get("convolution_mode", "full")

        if fourier_shift_scores:
            temp = backend.roll(temp, shift=fourier_shift, axis=axis)

        score = apply_convolution_mode(
            temp, convolution_mode=convolution_mode, s1=targetshape, s2=templateshape
        )
        if callback_class is not None:
            callback(
                score,
                rotation_matrix=rotation,
                rotation_index=index,
                **callback_class_args,
            )

    return callback


def device_memory_handler(func: Callable):
    """Decorator function providing SharedMemory Handler."""

    @wraps(func)
    def inner_function(*args, **kwargs):
        return_value = None
        last_type, last_value, last_traceback = sys.exc_info()
        try:
            with SharedMemoryManager() as smh:
                with backend.set_device(kwargs.get("gpu_index", 0)):
                    return_value = func(shared_memory_handler=smh, *args, **kwargs)
        except Exception as e:
            print(e)
            last_type, last_value, last_traceback = sys.exc_info()
        finally:
            handle_traceback(last_type, last_value, last_traceback)
        return return_value

    return inner_function


@device_memory_handler
def scan(
    matching_data: MatchingData,
    matching_setup: Callable,
    matching_score: Callable,
    n_jobs: int = 4,
    callback_class: CallbackClass = None,
    callback_class_args: Dict = {},
    fftargs: Dict = {},
    pad_fourier: bool = True,
    interpolation_order: int = 3,
    jobs_per_callback_class: int = 8,
    **kwargs,
) -> Tuple:
    """
    Perform template matching between target and template and sample
    different rotations of template.

    Parameters
    ----------
    matching_data : MatchingData
        Template matching data.
    matching_setup : Callable
        Function pointer to setup function.
    matching_score : Callable
        Function pointer to scoring function.
    n_jobs : int, optional
        Number of parallel jobs. Default is 4.
    callback_class : type, optional
        Analyzer class pointer to operate on computed scores.
    callback_class_args : dict, optional
        Arguments passed to the callback_class. Default is an empty dictionary.
    fftargs : dict, optional
        Arguments for the FFT operations. Default is an empty dictionary.
    pad_fourier: bool, optional
        Whether to pad target and template to the full convolution shape.
    interpolation_order : int, optional
        Order of spline interpolation for rotations.
    jobs_per_callback_class : int, optional
        How many jobs should be processed by a single callback_class instance,
        if ones is provided.
    **kwargs : various
        Additional arguments.

    Returns
    -------
    Tuple
        The merged results from callback_class if provided otherwise None.
    """
    matching_data.to_backend()
    fourier_pad = matching_data._templateshape
    fourier_shift = backend.zeros(len(fourier_pad))
    if not pad_fourier:
        fourier_pad = backend.full(shape=fourier_shift.shape, fill_value=1, dtype=int)

    convolution_shape, fast_shape, fast_ft_shape = backend.compute_convolution_shapes(
        matching_data._target.shape, fourier_pad
    )
    if not pad_fourier:
        fourier_shift = 1 - backend.astype(
            backend.divide(matching_data._templateshape, 2), int
        )
        fourier_shift -= backend.mod(matching_data._templateshape, 2)
        fourier_shift = backend.flip(fourier_shift, axis=(0,))
        shape_diff = backend.subtract(fast_shape, convolution_shape)
        shape_diff = backend.astype(backend.divide(shape_diff, 2), int)
        backend.add(fourier_shift, shape_diff, out=fourier_shift)

    callback_class_args["fourier_shift"] = fourier_shift
    rfftn, irfftn = backend.build_fft(
        fast_shape=fast_shape,
        fast_ft_shape=fast_ft_shape,
        real_dtype=matching_data._default_dtype,
        complex_dtype=matching_data._complex_dtype,
        fftargs=fftargs,
    )
    setup = matching_setup(
        rfftn=rfftn,
        irfftn=irfftn,
        template=matching_data.template,
        template_mask=matching_data.template_mask,
        target=matching_data.target,
        target_mask=matching_data.target_mask,
        fast_shape=fast_shape,
        fast_ft_shape=fast_ft_shape,
        real_dtype=matching_data._default_dtype,
        complex_dtype=matching_data._complex_dtype,
        callback_class=callback_class,
        callback_class_args=callback_class_args,
        **kwargs,
    )
    rfftn, irfftn = None, None

    template_filter, preprocessor = None, Preprocessor()
    for method, parameters in matching_data.template_filter.items():
        parameters["shape"] = fast_shape
        parameters["omit_negative_frequencies"] = True
        out = preprocessor.apply_method(method=method, parameters=parameters)
        if template_filter is None:
            template_filter = out
        np.multiply(template_filter, out, out=template_filter)

    if template_filter is None:
        template_filter = backend.full(
            shape=(1,), fill_value=1, dtype=backend._default_dtype
        )
    else:
        template_filter = backend.to_backend_array(template_filter)

    template_filter = backend.astype(template_filter, backend._default_dtype)
    template_filter_buffer = backend.arr_to_sharedarr(
        arr=template_filter,
        shared_memory_handler=kwargs.get("shared_memory_handler", None),
    )
    setup["template_filter"] = (
        template_filter_buffer,
        template_filter.shape,
        template_filter.dtype,
    )

    callback_class_args["translation_offset"] = backend.astype(
        matching_data._translation_offset, int
    )
    callback_class_args["thread_safe"] = n_jobs > 1
    callback_class_args["gpu_index"] = kwargs.get("gpu_index", -1)

    n_callback_classes = max(n_jobs // jobs_per_callback_class, 1)
    callback_class = setup.pop("callback_class", callback_class)
    callback_class_args = setup.pop("callback_class_args", callback_class_args)
    callback_classes = [callback_class for _ in range(n_callback_classes)]
    if callback_class == MaxScoreOverRotations:
        score_space_shape = backend.subtract(
            matching_data.target.shape,
            matching_data._target_pad,
        )
        callback_classes = [
            class_name(
                score_space_shape=score_space_shape,
                score_space_dtype=matching_data._default_dtype,
                shared_memory_handler=kwargs.get("shared_memory_handler", None),
                rotation_space_dtype=backend._default_dtype_int,
                **callback_class_args,
            )
            for class_name in callback_classes
        ]

    matching_data._target, matching_data._template = None, None
    matching_data._target_mask, matching_data._template_mask = None, None

    setup["fftargs"] = fftargs.copy()
    convolution_mode = "same"
    if backend.sum(matching_data._target_pad) > 0:
        convolution_mode = "valid"
    setup["convolution_mode"] = convolution_mode
    setup["interpolation_order"] = interpolation_order
    rotation_list = matching_data._split_rotations_on_jobs(n_jobs)

    backend.free_cache()

    def _run_scoring(backend_name, backend_args, rotations, **kwargs):
        from tme.backends import backend

        backend.change_backend(backend_name, **backend_args)
        return matching_score(rotations=rotations, **kwargs)

    callbacks = Parallel(n_jobs=n_jobs)(
        delayed(_run_scoring)(
            backend_name=backend._backend_name,
            backend_args=backend._backend_args,
            rotations=rotation,
            callback_class=callback_classes[index % n_callback_classes],
            callback_class_args=callback_class_args,
            **setup,
        )
        for index, rotation in enumerate(rotation_list)
    )

    callbacks = [
        tuple(callback)
        for callback in callbacks[0:n_callback_classes]
        if callback is not None
    ]
    backend.free_cache()

    merged_callback = None
    if callback_class is not None:
        merged_callback = callback_class.merge(
            callbacks,
            **callback_class_args,
            score_indices=matching_data.indices,
            inner_merge=True,
        )

    return merged_callback


def scan_subsets(
    matching_data: MatchingData,
    matching_score: Callable,
    matching_setup: Callable,
    callback_class: CallbackClass = None,
    callback_class_args: Dict = {},
    job_schedule: Tuple[int] = (1, 1),
    target_splits: Dict = {},
    template_splits: Dict = {},
    pad_target_edges: bool = False,
    pad_fourier: bool = True,
    interpolation_order: int = 3,
    jobs_per_callback_class: int = 8,
    **kwargs,
) -> Tuple:
    """
    Wrapper around :py:meth:`scan` that supports template matching on splits
    of template and target.

    Parameters
    ----------
    matching_data : MatchingData
        Template matching data.
    matching_func : type
        Function pointer to setup function.
    matching_score : type
        Function pointer to scoring function.
    callback_class : type, optional
        Analyzer class pointer to operate on computed scores.
    callback_class_args : dict, optional
        Arguments passed to the callback_class. Default is an empty dictionary.
    job_schedule : tuple of int, optional
        Schedule of jobs. Default is (1, 1).
    target_splits : dict, optional
        Splits for target. Default is an empty dictionary, i.e. no splits
    template_splits : dict, optional
        Splits for template. Default is an empty dictionary, i.e. no splits.
    pad_target_edges : bool, optional
        Whether to pad the target boundaries by half the template shape
        along each axis.
    pad_fourier: bool, optional
        Whether to pad target and template to the full convolution shape.
    interpolation_order : int, optional
        Order of spline interpolation for rotations.
    jobs_per_callback_class : int, optional
        How many jobs should be processed by a single callback_class instance,
        if ones is provided.
    **kwargs : various
        Additional arguments.

    Notes
    -----
        Objects in matching_data might be destroyed during computation.

    Returns
    -------
    Tuple
        The merged results from callback_class if provided otherwise None.
    """
    target_splits = split_numpy_array_slices(
        matching_data.target.shape, splits=target_splits
    )
    template_splits = split_numpy_array_slices(
        matching_data._templateshape, splits=template_splits
    )

    target_pad = np.zeros(len(matching_data.target.shape), dtype=int)
    if pad_target_edges:
        target_pad = np.subtract(
            matching_data._templateshape, np.mod(matching_data._templateshape, 2)
        )
    outer_jobs, inner_jobs = job_schedule
    results = Parallel(n_jobs=outer_jobs)(
        delayed(_run_inner)(
            backend_name=backend._backend_name,
            backend_args=backend._backend_args,
            matching_data=matching_data.subset_by_slice(
                target_slice=target_split,
                target_pad=target_pad,
                template_slice=template_split,
            ),
            matching_score=matching_score,
            matching_setup=matching_setup,
            n_jobs=inner_jobs,
            callback_class=callback_class,
            callback_class_args=callback_class_args,
            interpolation_order=interpolation_order,
            pad_fourier=pad_fourier,
            gpu_index=index % outer_jobs,
            **kwargs,
        )
        for index, (target_split, template_split) in enumerate(
            product(target_splits, template_splits)
        )
    )

    matching_data._target, matching_data._template = None, None
    matching_data._target_mask, matching_data._template_mask = None, None

    if callback_class is not None:
        candidates = callback_class.merge(
            results, **callback_class_args, inner_merge=False
        )
        return candidates


MATCHING_EXHAUSTIVE_REGISTER = {
    "CC": (cc_setup, corr_scoring),
    "LCC": (lcc_setup, corr_scoring),
    "CORR": (corr_setup, corr_scoring),
    "CAM": (cam_setup, corr_scoring),
    "FLCSphericalMask": (flcSphericalMask_setup, corr_scoring),
    "FLC": (flc_setup, flc_scoring),
    "MCC": (mcc_setup, mcc_scoring),
}


def register_matching_exhaustive(
    matching: str,
    matching_setup: Callable,
    matching_scoring: Callable,
    memory_class: MatchingMemoryUsage,
) -> None:
    """
    Registers a new matching scheme.

    Parameters
    ----------
    matching : str
        Name of the matching method.
    matching_setup : Callable
        The setup function associated with the name.
    matching_scoring : Callable
        The scoring function associated with the name.
    memory_class : MatchingMemoryUsage
        The custom memory estimation class extending
        :py:class:`tme.matching_memory.MatchingMemoryUsage`.

    Raises
    ------
    ValueError
        If a function with the name ``matching`` already exists in the registry.
    ValueError
        If ``memory_class`` is not a subclass of
        :py:class:`tme.matching_memory.MatchingMemoryUsage`.
    """

    if matching in MATCHING_EXHAUSTIVE_REGISTER:
        raise ValueError(f"A method with name '{matching}' is already registered.")
    if not issubclass(memory_class, MatchingMemoryUsage):
        raise ValueError(f"{memory_class} is not a subclass of {MatchingMemoryUsage}.")

    MATCHING_EXHAUSTIVE_REGISTER[matching] = (matching_setup, matching_scoring)
    MATCHING_MEMORY_REGISTRY[matching] = memory_class



#!python3
""" Simplify picking adequate filtering and masking parameters using a GUI.
    Exposes tme.preprocessor.Preprocessor and tme.fitter_utils member functions
    to achieve this aim.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import inspect
from typing import Tuple, Callable, List
from typing_extensions import Annotated

import numpy as np
import pandas as pd
import napari
from napari.layers import Image
from napari.utils.events import EventedList

from magicgui import widgets
from qtpy.QtWidgets import QFileDialog
from numpy.typing import NDArray

from tme import Preprocessor, Density
from tme.matching_utils import create_mask

preprocessor = Preprocessor()
SLIDER_MIN, SLIDER_MAX = 0, 25


def gaussian_filter(template, sigma: float, **kwargs: dict):
    return preprocessor.gaussian_filter(template=template, sigma=sigma, **kwargs)


def bandpass_filter(
    template,
    minimum_frequency: float,
    maximum_frequency: float,
    gaussian_sigma: float,
    **kwargs: dict,
):
    return preprocessor.bandpass_filter(
        template=template,
        minimum_frequency=minimum_frequency,
        maximum_frequency=maximum_frequency,
        sampling_rate=1,
        gaussian_sigma=gaussian_sigma,
        **kwargs,
    )


def difference_of_gaussian_filter(
    template, sigmas: Tuple[float, float], **kwargs: dict
):
    low_sigma, high_sigma = sigmas
    return preprocessor.difference_of_gaussian_filter(
        template=template, low_sigma=low_sigma, high_sigma=high_sigma, **kwargs
    )


def edge_gaussian_filter(
    template,
    sigma: float,
    edge_algorithm: Annotated[
        str,
        {"choices": ["sobel", "prewitt", "laplace", "gaussian", "gaussian_laplace"]},
    ],
    reverse: bool = False,
    **kwargs: dict,
):
    return preprocessor.edge_gaussian_filter(
        template=template,
        sigma=sigma,
        reverse=reverse,
        edge_algorithm=edge_algorithm,
    )


def local_gaussian_filter(
    template,
    lbd: float,
    sigma_range: Tuple[float, float],
    gaussian_sigma: float,
    reverse: bool = False,
    **kwargs: dict,
):
    return preprocessor.local_gaussian_filter(
        template=template,
        lbd=lbd,
        sigma_range=sigma_range,
        gaussian_sigma=gaussian_sigma,
    )


def ntree(
    template,
    sigma_range: Tuple[float, float],
    **kwargs: dict,
):
    return preprocessor.ntree_filter(template=template, sigma_range=sigma_range)


def mean(
    template,
    width: int,
    **kwargs: dict,
):
    return preprocessor.mean_filter(template=template, width=width)


def wedge(
    template: NDArray,
    tilt_start: float,
    tilt_stop: float,
    gaussian_sigma: float,
    tilt_axis: int = 1,
    infinite_plane: bool = True,
    extrude_plane: bool = True,
):
    template_ft = np.fft.rfftn(template)
    wedge_mask = preprocessor.continuous_wedge_mask(
        start_tilt=tilt_start,
        stop_tilt=tilt_stop,
        tilt_axis=tilt_axis,
        shape=template.shape,
        sigma=gaussian_sigma,
        omit_negative_frequencies=True,
        infinite_plane=infinite_plane,
        extrude_plane=extrude_plane,
    )
    np.multiply(template_ft, wedge_mask, out=template_ft)
    template = np.real(np.fft.irfftn(template_ft))
    return template


def widgets_from_function(function: Callable, exclude_params: List = ["self"]):
    """
    Creates list of magicui widgets by inspecting function typing ann
    """
    ret = []
    for name, param in inspect.signature(function).parameters.items():
        if name in exclude_params:
            continue

        if param.annotation is float:
            widget = widgets.FloatSpinBox(
                name=name,
                value=param.default if param.default != inspect._empty else 0,
                min=SLIDER_MIN,
                step=0.5,
            )
        elif param.annotation == Tuple[float, float]:
            widget = widgets.FloatRangeSlider(
                name=param.name,
                value=param.default
                if param.default != inspect._empty
                else (0.0, SLIDER_MAX / 2),
                min=SLIDER_MIN,
                max=SLIDER_MAX,
            )
        elif param.annotation is int:
            widget = widgets.SpinBox(
                name=name,
                value=param.default if param.default != inspect._empty else 0,
            )
        elif param.annotation is bool:
            widget = widgets.CheckBox(
                name=name,
                value=param.default if param.default != inspect._empty else False,
            )
        elif hasattr(param.annotation, "__metadata__"):
            metadata = param.annotation.__metadata__[0]
            if "choices" in metadata:
                widget = widgets.ComboBox(
                    name=param.name,
                    choices=metadata["choices"],
                    value=param.default
                    if param.default != inspect._empty
                    else metadata["choices"][0],
                )
        else:
            continue
        ret.append(widget)
    return ret


WRAPPED_FUNCTIONS = {
    "gaussian_filter": gaussian_filter,
    "bandpass_filter": bandpass_filter,
    "edge_gaussian_filter": edge_gaussian_filter,
    "ntree_filter": ntree,
    "local_gaussian_filter": local_gaussian_filter,
    "difference_of_gaussian_filter": difference_of_gaussian_filter,
    "mean_filter": mean,
    "continuous_wedge_mask": wedge,
}

EXCLUDED_FUNCTIONS = [
    "apply_method",
    "method_to_id",
    "wedge_mask",
    "fourier_crop",
    "fourier_uncrop",
    "interpolate_box",
    "molmap",
    "local_gaussian_alignment_filter",
    # "continuous_wedge_mask",
    "wedge_mask",
    "bandpass_mask",
]


class FilterWidget(widgets.Container):
    def __init__(self, preprocessor, viewer):
        super().__init__(layout="vertical")

        self.preprocessor = preprocessor
        self.viewer = viewer
        self.name_mapping = {}
        self.action_widgets = []

        self.layer_dropdown = widgets.ComboBox(
            name="Target Layer", choices=self._get_layer_names()
        )
        self.append(self.layer_dropdown)
        self.viewer.layers.events.inserted.connect(self._update_layer_dropdown)
        self.viewer.layers.events.removed.connect(self._update_layer_dropdown)

        self.method_dropdown = widgets.ComboBox(
            name="Choose Filter", choices=self._get_method_names()
        )
        self.method_dropdown.changed.connect(self._on_method_changed)
        self.append(self.method_dropdown)

        self.apply_btn = widgets.PushButton(text="Apply Filter", enabled=False)
        self.apply_btn.changed.connect(self._action)
        self.append(self.apply_btn)

        # Create GUI for initially selected filtering method
        self._on_method_changed(None)

    def _get_method_names(self):
        method_names = [
            name
            for name, member in inspect.getmembers(self.preprocessor, inspect.ismethod)
            if not name.startswith("_") and name not in EXCLUDED_FUNCTIONS
        ]

        sanitized_names = [self._sanitize_name(name) for name in method_names]
        self.name_mapping.update(dict(zip(sanitized_names, method_names)))

        return sanitized_names

    def _sanitize_name(self, name: str) -> str:
        # Replace underscores with spaces and capitalize each word
        removes = ["blur", "filter"]
        for remove in removes:
            name = name.replace(remove, "")
        return name.strip().replace("_", " ").title()

    def _desanitize_name(self, name: str) -> str:
        name = name.lower().strip()
        for function_name, _ in inspect.getmembers(self.preprocessor, inspect.ismethod):
            if function_name.startswith(name):
                return function_name
        return name

    def _get_function(self, name: str):
        function = WRAPPED_FUNCTIONS.get(name, None)
        if not function:
            function = getattr(self.preprocessor, name, None)
        return function

    def _on_method_changed(self, event=None):
        # Clear previous parameter widgets
        for widget in self.action_widgets:
            self.remove(widget)
        self.action_widgets.clear()

        function_name = self.name_mapping.get(self.method_dropdown.value)
        function = self._get_function(function_name)

        widgets = widgets_from_function(function, exclude_params=["self", "template"])
        for widget in widgets:
            self.action_widgets.append(widget)
            self.insert(-1, widget)

    def _update_layer_dropdown(self, event: EventedList):
        """Update the dropdown menu when layers change."""
        self.layer_dropdown.choices = self._get_layer_names()
        self.apply_btn.enabled = bool(self.viewer.layers)

    def _get_layer_names(self):
        """Return list of layer names in the viewer."""
        return sorted([layer.name for layer in self.viewer.layers])

    def _action(self, event):
        selected_layer = self.viewer.layers[self.layer_dropdown.value]
        selected_layer_metadata = selected_layer.metadata.copy()
        kwargs = {widget.name: widget.value for widget in self.action_widgets}

        function_name = self.name_mapping.get(self.method_dropdown.value)
        function = self._get_function(function_name)

        processed_data = function(selected_layer.data, **kwargs)

        new_layer_name = f"{selected_layer.name} ({self.method_dropdown.value})"

        if new_layer_name in self.viewer.layers:
            selected_layer = self.viewer.layers[new_layer_name]

        filter_name = self._desanitize_name(self.method_dropdown.value)
        used_filter = selected_layer.metadata.get("used_filter", False)
        if used_filter == filter_name:
            selected_layer.data = processed_data
        else:
            new_layer = self.viewer.add_image(
                data=processed_data,
                name=new_layer_name,
            )
            metadata = selected_layer_metadata.copy()
            if "filter_parameters" not in metadata:
                metadata["filter_parameters"] = []
            metadata["filter_parameters"].append({filter_name: kwargs.copy()})
            metadata["used_filter"] = filter_name
            new_layer.metadata = metadata


def sphere_mask(
    template: NDArray, center_x: float, center_y: float, center_z: float, radius: float
):
    return create_mask(
        mask_type="ellipse",
        shape=template.shape,
        center=(center_x, center_y, center_z),
        radius=radius,
    )


def ellipsod_mask(
    template: NDArray,
    center_x: float,
    center_y: float,
    center_z: float,
    radius_x: float,
    radius_y: float,
    radius_z: float,
):
    return create_mask(
        mask_type="ellipse",
        shape=template.shape,
        center=(center_x, center_y, center_z),
        radius=(radius_x, radius_y, radius_z),
    )


def box_mask(
    template: NDArray,
    center_x: float,
    center_y: float,
    center_z: float,
    height_x: int,
    height_y: int,
    height_z: int,
):
    return create_mask(
        mask_type="box",
        shape=template.shape,
        center=(center_x, center_y, center_z),
        height=(height_x, height_y, height_z),
    )


def tube_mask(
    template: NDArray,
    symmetry_axis: int,
    center_x: float,
    center_y: float,
    center_z: float,
    inner_radius: float,
    outer_radius: float,
    height: int,
):
    return create_mask(
        mask_type="tube",
        shape=template.shape,
        symmetry_axis=symmetry_axis,
        base_center=(center_x, center_y, center_z),
        inner_radius=inner_radius,
        outer_radius=outer_radius,
        height=height,
    )


def wedge_mask(
    template: NDArray,
    tilt_start: float,
    tilt_stop: float,
    gaussian_sigma: float,
    tilt_axis: int = 1,
    omit_negative_frequencies: bool = True,
    extrude_plane: bool = True,
    infinite_plane: bool = True,
):
    wedge_mask = preprocessor.continuous_wedge_mask(
        start_tilt=tilt_start,
        stop_tilt=tilt_stop,
        tilt_axis=tilt_axis,
        shape=template.shape,
        sigma=gaussian_sigma,
        omit_negative_frequencies=omit_negative_frequencies,
        extrude_plane=extrude_plane,
        infinite_plane=infinite_plane,
    )
    wedge_mask = np.fft.fftshift(wedge_mask)
    return wedge_mask

def threshold_mask(
    template: NDArray,
    standard_deviation : float = 5.0,
    invert : bool = False
):
    template_mean = template.mean()
    template_deviation = standard_deviation * template.std()
    upper = template_mean + template_deviation
    lower = template_mean - template_deviation
    mask = np.logical_and(template > lower, template < upper)
    if invert:
        np.invert(mask, out = mask)

    return mask



class MaskWidget(widgets.Container):
    def __init__(self, viewer):
        super().__init__(layout="vertical")

        self.viewer = viewer
        self.action_widgets = []

        self.action_button = widgets.PushButton(text="Create mask", enabled=False)
        self.action_button.changed.connect(self._action)

        self.methods = {
            "Sphere": sphere_mask,
            "Ellipsoid": ellipsod_mask,
            "Tube": tube_mask,
            "Box": box_mask,
            "Wedge": wedge_mask,
            "Threshold" : threshold_mask
        }

        self.method_dropdown = widgets.ComboBox(
            name="Choose Mask", choices=list(self.methods.keys())
        )
        self.method_dropdown.changed.connect(self._on_method_changed)

        self.adapt_button = widgets.PushButton(text="Adapt to layer", enabled=False)
        self.adapt_button.changed.connect(self._update_initial_values)

        self.viewer.layers.selection.events.active.connect(
            self._update_action_button_state
        )

        self.align_button = widgets.PushButton(text="Align to axis", enabled=False)
        self.align_button.changed.connect(self._align_with_axis)
        self.density_field = widgets.Label()
        # self.density_field.value = f"Positive Density in Mask: {0:.2f}%"

        self.append(self.method_dropdown)
        self.append(self.adapt_button)
        self.append(self.align_button)
        self.append(self.action_button)
        self.append(self.density_field)

        # Create GUI for initially selected filtering method
        self._on_method_changed(None)

    def _update_action_button_state(self, event):
        self.align_button.enabled = bool(self.viewer.layers.selection.active)
        self.action_button.enabled = bool(self.viewer.layers.selection.active)
        self.adapt_button.enabled = bool(self.viewer.layers.selection.active)

    def _align_with_axis(self):
        active_layer = self.viewer.layers.selection.active

        if active_layer.metadata.get("is_aligned", False):
            return

        coords = np.array(np.where(active_layer.data > 0)).T
        centered_coords = coords - np.mean(coords, axis=0)
        cov_matrix = np.cov(centered_coords, rowvar=False)

        eigenvalues, eigenvectors = np.linalg.eigh(cov_matrix)
        principal_eigenvector = eigenvectors[:, np.argmax(eigenvalues)]

        rotation_axis = np.cross(principal_eigenvector, [1, 0, 0])
        rotation_angle = np.arccos(np.dot(principal_eigenvector, [1, 0, 0]))
        k = rotation_axis / np.linalg.norm(rotation_axis)
        K = np.array([[0, -k[2], k[1]], [k[2], 0, -k[0]], [-k[1], k[0], 0]])
        rotation_matrix = np.eye(3)
        rotation_matrix += np.sin(rotation_angle) * K
        rotation_matrix += (1 - np.cos(rotation_angle)) * np.dot(K, K)

        rotated_data = Density.rotate_array(
            arr=active_layer.data,
            rotation_matrix=rotation_matrix,
            use_geometric_center=False,
        )
        eps = np.finfo(rotated_data.dtype).eps
        rotated_data[rotated_data < eps] = 0

        active_layer.metadata["is_aligned"] = True
        active_layer.data = rotated_data

    def _update_initial_values(self, event=None):
        active_layer = self.viewer.layers.selection.active
        center_of_mass = Density.center_of_mass(np.abs(active_layer.data), 0)
        coordinates = np.array(np.where(active_layer.data > 0))
        coordinates_min = coordinates.min(axis=1)
        coordinates_max = coordinates.max(axis=1)
        coordinates_heights = coordinates_max - coordinates_min
        coordinate_radius = np.divide(coordinates_heights, 2)
        center_of_mass = coordinate_radius + coordinates_min

        defaults = dict(zip(["center_x", "center_y", "center_z"], center_of_mass))
        defaults.update(
            dict(zip(["radius_x", "radius_y", "radius_z"], coordinate_radius))
        )
        defaults.update(
            dict(zip(["height_x", "height_y", "height_z"], coordinates_heights))
        )

        defaults["radius"] = np.min(coordinate_radius)
        defaults["inner_radius"] = np.min(coordinate_radius)
        defaults["outer_radius"] = np.max(coordinate_radius)
        defaults["height"] = defaults["radius"]

        for widget in self.action_widgets:
            if widget.name in defaults:
                widget.value = defaults[widget.name]

    def _on_method_changed(self, event=None):
        for widget in self.action_widgets:
            self.remove(widget)
        self.action_widgets.clear()

        function = self.methods.get(self.method_dropdown.value)
        widgets = widgets_from_function(function)
        for widget in widgets:
            self.action_widgets.append(widget)
            self.insert(1, widget)

    def _action(self):
        function = self.methods.get(self.method_dropdown.value)

        selected_layer = self.viewer.layers.selection.active
        kwargs = {widget.name: widget.value for widget in self.action_widgets}
        processed_data = function(template=selected_layer.data, **kwargs)
        new_layer_name = f"{selected_layer.name} ({self.method_dropdown.value})"

        if new_layer_name in self.viewer.layers:
            selected_layer = self.viewer.layers[new_layer_name]

        processed_data = processed_data.astype(np.float32)
        metadata = selected_layer.metadata
        mask = metadata.get("mask", False)
        if mask == self.method_dropdown.value:
            selected_layer.data = processed_data
        else:
            new_layer = self.viewer.add_image(
                data=processed_data,
                name=new_layer_name,
            )
            metadata = selected_layer.metadata.copy()
            metadata["filter_parameters"] = {self.method_dropdown.value: kwargs.copy()}
            metadata["mask"] = self.method_dropdown.value
            metadata["origin_layer"] = selected_layer.name
            new_layer.metadata = metadata

        origin_layer = metadata["origin_layer"]
        if origin_layer in self.viewer.layers:
            origin_layer = self.viewer.layers[origin_layer]
            if np.allclose(origin_layer.data.shape, processed_data.shape):
                in_mask = np.sum(np.fmax(origin_layer.data * processed_data, 0))
                in_mask /= np.sum(np.fmax(origin_layer.data, 0))
                in_mask *= 100
                self.density_field.value = f"Positive Density in Mask: {in_mask:.2f}%"


class ExportWidget(widgets.Container):
    def __init__(self, viewer):
        super().__init__(layout="vertical")

        self.viewer = viewer
        self.selected_filename = ""

        horizontal_container = widgets.Container(layout="horizontal")

        self.gzip_output = widgets.CheckBox(name="gzip", value=False, label="gzip")
        self.export_button = widgets.PushButton(name="Export", text="Export")
        self.export_button.clicked.connect(self._get_save_path)

        horizontal_container.append(self.export_button)
        horizontal_container.append(self.gzip_output)

        self.append(horizontal_container)

        self.export_button.enabled = bool(self.viewer.layers.selection.active)
        self.viewer.layers.selection.events.active.connect(
            self._update_export_button_state
        )

    def _get_save_path(self, event):
        options = QFileDialog.Options()
        path, _ = QFileDialog.getSaveFileName(
            self.native,
            "Save As...",
            "",
            "MRC Files (*.mrc)",
            options=options,
        )
        if path:
            self.selected_filename = path
            self._export_data()

    def _update_export_button_state(self, event):
        """Update the enabled state of the export button based on the active layer."""
        self.export_button.enabled = bool(self.viewer.layers.selection.active)

    def _export_data(self):
        selected_layer = self.viewer.layers.selection.active
        if selected_layer and isinstance(selected_layer, Image):
            selected_layer.metadata["write_gzip"] = self.gzip_output.value
            selected_layer.save(path=self.selected_filename)


class PointCloudWidget(widgets.Container):
    def __init__(self, viewer):
        super().__init__(layout="vertical")

        self.viewer = viewer
        self.dataframes = {}

        self.import_button = widgets.PushButton(
            name="Import", text="Import Point Cloud"
        )
        self.import_button.clicked.connect(self._get_load_path)

        self.export_button = widgets.PushButton(
            name="Export", text="Export Point Cloud"
        )
        self.export_button.clicked.connect(self._export_point_cloud)
        self.export_button.enabled = False

        self.append(self.import_button)
        self.append(self.export_button)
        self.viewer.layers.selection.events.changed.connect(self._update_buttons)

    def _update_buttons(self, event):
        is_pointcloud = isinstance(
            self.viewer.layers.selection.active, napari.layers.Points
        )
        if self.viewer.layers.selection.active and is_pointcloud:
            self.export_button.enabled = True
        else:
            self.export_button.enabled = False

    def _export_point_cloud(self, event):
        options = QFileDialog.Options()
        filename, _ = QFileDialog.getSaveFileName(
            self.native,
            "Save Point Cloud File...",
            "",
            "TSV Files (*.tsv);;All Files (*)",
            options=options,
        )

        if filename:
            layer = self.viewer.layers.selection.active
            if layer and isinstance(layer, napari.layers.Points):
                original_dataframe = self.dataframes.get(layer.name, pd.DataFrame())

                export_data = pd.DataFrame(layer.data, columns=["z", "y", "x"])
                merged_data = pd.merge(
                    export_data, original_dataframe, on=["z", "y", "x"], how="left"
                )
                merged_data.to_csv(filename, sep="\t", index=False)

    def _get_load_path(self, event):
        options = QFileDialog.Options()
        filename, _ = QFileDialog.getOpenFileName(
            self.native,
            "Open Point Cloud File...",
            "",
            "TSV Files (*.tsv);;All Files (*)",
            options=options,
        )
        if filename:
            self._load_point_cloud(filename)

    def _load_point_cloud(self, filename):
        dataframe = pd.read_csv(filename, sep="\t")
        points = dataframe[["z", "y", "x"]].values
        layer_name = filename.split("/")[-1]
        self.viewer.add_points(points, size=10, name=layer_name)
        self.dataframes[layer_name] = dataframe


def main():
    viewer = napari.Viewer()

    filter_widget = FilterWidget(preprocessor, viewer)
    mask_widget = MaskWidget(viewer)
    export_widget = ExportWidget(viewer)
    point_cloud = PointCloudWidget(viewer)

    viewer.window.add_dock_widget(widget=filter_widget, name="Preprocess", area="right")
    viewer.window.add_dock_widget(widget=mask_widget, name="Mask", area="right")
    viewer.window.add_dock_widget(widget=point_cloud, name="PointCloud", area="left")

    viewer.window.add_dock_widget(widget=export_widget, name="Export", area="right")

    napari.run()


if __name__ == "__main__":
    main()



""" Backend using cupy and GPU acceleration for
    template matching.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import warnings
from typing import Tuple, Dict, Callable
from contextlib import contextmanager

import numpy as np
from numpy.typing import NDArray

from .npfftw_backend import NumpyFFTWBackend
from ..types import CupyArray


class CupyBackend(NumpyFFTWBackend):
    """
    A cupy based backend for template matching
    """

    def __init__(
        self, default_dtype=None, complex_dtype=None, default_dtype_int=None, **kwargs
    ):
        import cupy as cp
        from cupyx.scipy.fft import get_fft_plan
        from cupyx.scipy.ndimage import affine_transform
        from cupyx.scipy.ndimage import maximum_filter

        default_dtype = cp.float32 if default_dtype is None else default_dtype
        complex_dtype = cp.complex64 if complex_dtype is None else complex_dtype
        default_dtype_int = cp.int32 if default_dtype_int is None else default_dtype_int

        super().__init__(
            array_backend=cp,
            default_dtype=default_dtype,
            complex_dtype=complex_dtype,
            default_dtype_int=default_dtype_int,
        )
        self.get_fft_plan = get_fft_plan
        self.affine_transform = affine_transform
        self.maximum_filter = maximum_filter

    def to_backend_array(self, arr: NDArray) -> CupyArray:
        if isinstance(arr, self._array_backend.ndarray):
            return arr
        return self._array_backend.asarray(arr)

    def to_numpy_array(self, arr: CupyArray) -> NDArray:
        return self._array_backend.asnumpy(arr)

    def to_cpu_array(self, arr: NDArray) -> NDArray:
        return self.to_numpy_array(arr)

    def sharedarr_to_arr(
        self, shape: Tuple[int], dtype: str, shm: CupyArray
    ) -> CupyArray:
        return shm

    @staticmethod
    def arr_to_sharedarr(
        arr: CupyArray, shared_memory_handler: type = None
    ) -> CupyArray:
        return arr

    def preallocate_array(self, shape: Tuple[int], dtype: type) -> NDArray:
        """
        Returns a byte-aligned array of zeros with specified shape and dtype.

        Parameters
        ----------
        shape : Tuple[int]
            Desired shape for the array.
        dtype : type
            Desired data type for the array.

        Returns
        -------
        NDArray
            Byte-aligned array of zeros with specified shape and dtype.
        """
        arr = self._array_backend.zeros(shape, dtype=dtype)
        return arr

    def unravel_index(self, indices, shape):
        return self._array_backend.unravel_index(indices=indices, dims=shape)

    def unique(self, ar, axis=None, *args, **kwargs):
        if axis is None:
            return self._array_backend.unique(ar=ar, axis=axis, *args, **kwargs)
        warnings.warn("Axis argument not yet supported in CupY, falling back to NumPy.")

        ret = np.unique(ar=self.to_numpy_array(ar), axis=axis, *args, **kwargs)
        if type(ret) != tuple:
            return self.to_backend_array(ret)
        return tuple(self.to_backend_array(k) for k in ret)

    def build_fft(
        self,
        fast_shape: Tuple[int],
        fast_ft_shape: Tuple[int],
        real_dtype: type,
        complex_dtype: type,
        fftargs: Dict = {},
        temp_real: NDArray = None,
        temp_fft: NDArray = None,
    ) -> Tuple[Callable, Callable]:
        """
        Build pyFFTW builder functions.

        Parameters
        ----------
        fast_shape : tuple
            Tuple of integers corresponding to fast convolution shape
            (see `compute_convolution_shapes`).
        fast_ft_shape : tuple
            Tuple of integers corresponding to the shape of the fourier
            transform array (see `compute_convolution_shapes`).
        real_dtype : dtype
            Numpy dtype of the inverse fourier transform.
        complex_dtype : dtype
            Numpy dtype of the fourier transform.
        fftargs : dict, optional
            Dictionary passed to pyFFTW builders.
        temp_real : NDArray, optional
            Temporary real numpy array, by default None.
        temp_fft : NDArray, optional
            Temporary fft numpy array, by default None.

        Returns
        -------
        tuple
            Tupple containing callable rfft and irfft object.
        """

        if temp_real is None:
            temp_real = self.preallocate_array(fast_shape, real_dtype)
        if temp_fft is None:
            temp_fft = self.preallocate_array(fast_ft_shape, complex_dtype)

        cache = self._array_backend.fft.config.get_plan_cache()
        cache.set_size(2)

        def rfftn(arr: CupyArray, out: CupyArray) -> None:
            out[:] = self.fft.rfftn(arr)[:]

        def irfftn(arr: CupyArray, out: CupyArray) -> None:
            out[:] = self.fft.irfftn(arr)[:]

        return rfftn, irfftn

    def compute_convolution_shapes(
        self, arr1_shape: Tuple[int], arr2_shape: Tuple[int]
    ) -> Tuple[Tuple[int], Tuple[int], Tuple[int]]:
        ret = super().compute_convolution_shapes(arr1_shape, arr2_shape)
        convolution_shape, fast_shape, fast_ft_shape = ret

        # cuFFT plans do not support automatic padding yet.
        is_odd = fast_shape[-1] % 2
        fast_shape[-1] += is_odd
        fast_ft_shape[-1] += is_odd

        return convolution_shape, fast_shape, fast_ft_shape

    def max_filter_coordinates(self, score_space, min_distance: Tuple[int]):
        score_box = tuple(min_distance for _ in range(score_space.ndim))
        max_filter = self.maximum_filter(score_space, size=score_box, mode="constant")
        max_filter = max_filter == score_space

        peaks = self._array_backend.array(self._array_backend.nonzero(max_filter)).T
        return peaks

    def rotate_array(
        self,
        arr: CupyArray,
        rotation_matrix: CupyArray,
        arr_mask: CupyArray = None,
        translation: CupyArray = None,
        use_geometric_center: bool = False,
        out: CupyArray = None,
        out_mask: CupyArray = None,
        order: int = 3,
    ) -> None:
        """
        Rotates coordinates of arr according to rotation_matrix.

        If no output array is provided, this method will compute an array with
        sufficient space to hold all elements. If both `arr` and `arr_mask`
        are provided, `arr_mask` will be centered according to arr.

        Parameters
        ----------
        arr : CupyArray
            The input array to be rotated.
        arr_mask : CupyArray, optional
            The mask of `arr` that will be equivalently rotated.
        rotation_matrix : CupyArray
            The rotation matrix to apply [d x d].
        translation : CupyArray
            The translation to apply [d].
        use_geometric_center : bool, optional
            Whether the rotation should be centered around the geometric
            or mass center. Default is mass center.
        out : CupyArray, optional
            The output array to write the rotation of `arr` to.
        out_mask : CupyArray, optional
            The output array to write the rotation of `arr_mask` to.
        order : int, optional
            Spline interpolation order. Has to be in the range 0-5.

        Notes
        -----
        Only a box of size arr, arr_mask will be consisdered for interpolation
        in out, out_mask.
        """

        rotate_mask = arr_mask is not None
        return_type = (out is None) + 2 * rotate_mask * (out_mask is None)
        translation = self.zeros(arr.ndim) if translation is None else translation

        center = self.divide(self.to_backend_array(arr.shape), 2)
        if not use_geometric_center:
            center = self.center_of_mass(arr, cutoff=0)

        rotation_matrix_inverted = self.linalg.inv(rotation_matrix)
        transformed_center = rotation_matrix_inverted @ center.reshape(-1, 1)
        transformed_center = transformed_center.reshape(-1)
        base_offset = self.subtract(center, transformed_center)
        offset = self.subtract(base_offset, translation)

        out = self.zeros_like(arr) if out is None else out
        out_slice = tuple(slice(0, stop) for stop in arr.shape)

        # Applying the prefilter leads to the creation of artifacts in the mask.
        self.affine_transform(
            input=arr,
            matrix=rotation_matrix_inverted,
            offset=offset,
            mode="constant",
            output=out[out_slice],
            order=order,
            prefilter=True,
        )

        if rotate_mask:
            out_mask = self.zeros_like(arr_mask) if out_mask is None else out_mask
            out_mask_slice = tuple(slice(0, stop) for stop in arr_mask.shape)
            self.affine_transform(
                input=arr_mask,
                matrix=rotation_matrix_inverted,
                offset=offset,
                mode="constant",
                output=out_mask[out_mask_slice],
                order=order,
                prefilter=False,
            )

        match return_type:
            case 0:
                return None
            case 1:
                return out
            case 2:
                return out_mask
            case 3:
                return out, out_mask

    def get_available_memory(self) -> int:
        with self._array_backend.cuda.Device():
            (
                free_memory,
                available_memory,
            ) = self._array_backend.cuda.runtime.memGetInfo()
        return free_memory

    @contextmanager
    def set_device(self, device_index: int):
        """
        Set the active GPU device as a context.

        This method sets the active GPU device for operations within the context.

        Parameters
        ----------
        device_index : int
            Index of the GPU device to be set as active.

        Yields
        ------
        None
            Operates as a context manager, yielding None and providing
            the set GPU context for enclosed operations.
        """
        with self._array_backend.cuda.Device(device_index):
            yield

    def device_count(self) -> int:
        """
        Return the number of available GPU devices.

        Returns
        -------
        int
            Number of available GPU devices.
        """
        return self._array_backend.cuda.runtime.getDeviceCount()



#!python3
""" CLI interface for basic pyTME template matching functions.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import os
import argparse
import warnings
import importlib.util
from sys import exit
from time import time
from copy import deepcopy
from os.path import abspath

import numpy as np

from tme import Density, Preprocessor, __version__
from tme.matching_utils import (
    get_rotation_matrices,
    compute_parallelization_schedule,
    euler_from_rotationmatrix,
    scramble_phases,
    generate_tempfile_name,
    write_pickle,
)
from tme.matching_exhaustive import scan_subsets, MATCHING_EXHAUSTIVE_REGISTER
from tme.matching_data import MatchingData
from tme.analyzer import (
    MaxScoreOverRotations,
    PeakCallerMaximumFilter,
)
from tme.backends import backend


def get_func_fullname(func) -> str:
    """Returns the full name of the given function, including its module."""
    return f"<function '{func.__module__}.{func.__name__}'>"


def print_block(name: str, data: dict, label_width=20) -> None:
    """Prints a formatted block of information."""
    print(f"\n> {name}")
    for key, value in data.items():
        formatted_value = str(value)
        print(f"  - {key + ':':<{label_width}} {formatted_value}")


def print_entry() -> None:
    width = 80
    text = f" pyTME v{__version__} "
    padding_total = width - len(text) - 2
    padding_left = padding_total // 2
    padding_right = padding_total - padding_left

    print("*" * width)
    print(f"*{ ' ' * padding_left }{text}{ ' ' * padding_right }*")
    print("*" * width)


def check_positive(value):
    ivalue = float(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive float." % value)
    return ivalue


def load_and_validate_mask(mask_target: "Density", mask_path: str, **kwargs):
    """
    Loadsa mask in CCP4/MRC format and assess whether the sampling_rate
    and shape matches its target.

    Parameters
    ----------
    mask_target : Density
        Object the mask should be applied to
    mask_path : str
        Path to the mask in CCP4/MRC format.
    kwargs : dict, optional
        Keyword arguments passed to :py:meth:`tme.density.Density.from_file`.
    Raise
    -----
    ValueError
        If shape or sampling rate do not match between mask_target and mask

    Returns
    -------
    Density
        A density instance if the mask was validated and loaded otherwise None
    """
    mask = mask_path
    if mask is not None:
        mask = Density.from_file(mask, **kwargs)
        mask.origin = deepcopy(mask_target.origin)
        if not np.allclose(mask.shape, mask_target.shape):
            raise ValueError(
                f"Expected shape of {mask_path} was {mask_target.shape},"
                f" got f{mask.shape}"
            )
        if not np.allclose(mask.sampling_rate, mask_target.sampling_rate):
            raise ValueError(
                f"Expected sampling_rate of {mask_path} was {mask_target.sampling_rate}"
                f", got f{mask.sampling_rate}"
            )
    return mask


def crop_data(data: Density, cutoff: float, data_mask: Density = None) -> bool:
    """
    Crop the provided data and mask to a smaller box based on a cutoff value.

    Parameters
    ----------
    data : Density
        The data that should be cropped.
    cutoff : float
        The threshold value to determine which parts of the data should be kept.
    data_mask : Density, optional
        A mask for the data that should be cropped.

    Returns
    -------
    bool
        Returns True if the data was adjusted (cropped), otherwise returns False.

    Notes
    -----
    Cropping is performed in place.
    """
    if cutoff is None:
        return False

    box = data.trim_box(cutoff=cutoff)
    box_mask = box
    if data_mask is not None:
        box_mask = data_mask.trim_box(cutoff=cutoff)
    box = tuple(
        slice(min(arr.start, mask.start), max(arr.stop, mask.stop))
        for arr, mask in zip(box, box_mask)
    )
    if box == tuple(slice(0, x) for x in data.shape):
        return False

    data.adjust_box(box)

    if data_mask:
        data_mask.adjust_box(box)

    return True


def parse_args():
    parser = argparse.ArgumentParser(description="Perform template matching.")
    parser.add_argument(
        "-m",
        "--target",
        dest="target",
        type=str,
        required=True,
        help="Path to a target in CCP4/MRC format.",
    )
    parser.add_argument(
        "--target_mask",
        dest="target_mask",
        type=str,
        required=False,
        help="Path to a mask for the target target in CCP4/MRC format.",
    )
    parser.add_argument(
        "--cutoff_target",
        dest="cutoff_target",
        type=float,
        required=False,
        help="Target contour level (used for cropping).",
        default=None,
    )
    parser.add_argument(
        "--cutoff_template",
        dest="cutoff_template",
        type=float,
        required=False,
        help="Template contour level (used for cropping).",
        default=None,
    )
    parser.add_argument(
        "-i",
        "--template",
        dest="template",
        type=str,
        required=True,
        help="Path to a template in PDB/MMCIF or CCP4/MRC format.",
    )
    parser.add_argument(
        "--template_mask",
        dest="template_mask",
        type=str,
        required=False,
        help="Path to a mask for the template in CCP4/MRC format.",
    )
    parser.add_argument(
        "-o",
        dest="output",
        type=str,
        required=False,
        default="output.pickle",
        help="Path to output pickle file.",
    )
    parser.add_argument(
        "-s",
        dest="score",
        type=str,
        default="FLCSphericalMask",
        help="Template matching scoring function.",
        choices=MATCHING_EXHAUSTIVE_REGISTER.keys(),
    )
    parser.add_argument(
        "-n",
        dest="cores",
        required=False,
        type=int,
        default=4,
        help="Number of cores used for template matching.",
    )
    parser.add_argument(
        "-r",
        "--ram",
        dest="memory",
        required=False,
        type=int,
        default=None,
        help="Amount of memory that can be used in bytes.",
    )
    parser.add_argument(
        "--memory_scaling",
        dest="memory_scaling",
        required=False,
        type=check_positive,
        default=0.85,
        help="Fraction of available memory that can be used."
        "Defaults to 0.85. Ignored if --ram is set",
    )
    parser.add_argument(
        "-a",
        dest="angular_sampling",
        type=check_positive,
        default=40.0,
        help="Angular sampling rate for template matching. "
        "A lower number yields more rotations.",
    )
    parser.add_argument(
        "-p",
        dest="peak_calling",
        action="store_true",
        default=False,
        help="When set perform peak calling instead of score aggregation.",
    )
    parser.add_argument(
        "--use_gpu",
        dest="use_gpu",
        action="store_true",
        default=False,
        help="Whether to perform computations on the GPU.",
    )
    parser.add_argument(
        "--gpu_indices",
        dest="gpu_indices",
        type=str,
        default=None,
        help="Comma-separated list of GPU indices to use. For example,"
        " 0,1 for the first and second GPU. Only used if --use_gpu is set."
        " If not provided but --use_gpu is set, CUDA_VISIBLE_DEVICES will"
        " be respected.",
    )
    parser.add_argument(
        "--invert_target_contrast",
        dest="invert_target_contrast",
        action="store_true",
        default=False,
        help="Invert the target contrast via multiplication with negative one and"
        " linear rescaling between zero and one. Note that this might lead to"
        " different baseline scores of individual target splits when using"
        " unnormalized scores. This option is intended for targets, where the"
        " object to-be-matched has negative values, i.e. tomograms.",
    )
    parser.add_argument(
        "--no_edge_padding",
        dest="no_edge_padding",
        action="store_true",
        default=False,
        help="Whether to pad the edges of the target. This is useful, if the target"
        " has a well defined bounding box, e.g. a density map.",
    )
    parser.add_argument(
        "--no_fourier_padding",
        dest="no_fourier_padding",
        action="store_true",
        default=False,
        help="Whether input arrays should be zero-padded to the full convolution shape"
        " for numerical stability. When working with very large targets such as"
        " tomograms it is safe to use this flag and benefit from the performance gain.",
    )
    parser.add_argument(
        "--scramble_phases",
        dest="scramble_phases",
        action="store_true",
        default=False,
        help="Whether to phase scramble the template for subsequent normalization.",
    )
    parser.add_argument(
        "--interpolation_order",
        dest="interpolation_order",
        required=False,
        type=int,
        default=3,
        help="Spline interpolation used during rotations. If less than zero"
        " no interpolation is performed.",
    )
    parser.add_argument(
        "--use_mixed_precision",
        dest="use_mixed_precision",
        action="store_true",
        default=False,
        help="Use float16 for real values operations where possible.",
    )
    parser.add_argument(
        "--use_memmap",
        dest="use_memmap",
        action="store_true",
        default=False,
        help="Use memmaps to offload large data objects to disk. This is"
        " particularly useful for large inputs when using --use_gpu..",
    )
    parser.add_argument(
        "--temp_directory",
        dest="temp_directory",
        default=None,
        help="Directory for temporary objects. Faster I/O typically improves runtime.",
    )
    parser.add_argument(
        "--gaussian_sigma",
        dest="gaussian_sigma",
        type=float,
        required=False,
        help="Sigma parameter for Gaussian filtering the template.",
    )
    parser.add_argument(
        "--bandpass_band",
        dest="bandpass_band",
        type=str,
        required=False,
        help="Comma separated start and stop frequency for bandpass filtering the"
        " template, e.g. 0.1, 0.5",
    )
    parser.add_argument(
        "--bandpass_smooth",
        dest="bandpass_smooth",
        type=float,
        required=False,
        default=None,
        help="Smooth parameter for the bandpass filter.",
    )

    parser.add_argument(
        "--tilt_range",
        dest="tilt_range",
        type=str,
        required=False,
        help="Comma separated start and stop stage tilt angle, e.g. '50,45'. Used"
        " to create a wedge mask to be applied to the template.",
    )
    parser.add_argument(
        "--tilt_step",
        dest="tilt_step",
        type=float,
        required=False,
        default=None,
        help="Step size between tilts, e.g. '5'. When set a more accurate"
        " wedge mask will be computed.",
    )
    parser.add_argument(
        "--wedge_smooth",
        dest="wedge_smooth",
        type=float,
        required=False,
        default=None,
        help="Gaussian sigma used to smooth the wedge mask.",
    )

    args = parser.parse_args()

    if args.interpolation_order < 0:
        args.interpolation_order = None

    if args.temp_directory is None:
        default = abspath(".")
        if os.environ.get("TMPDIR", None) is not None:
            default = os.environ.get("TMPDIR")
        args.temp_directory = default

    os.environ["TMPDIR"] = args.temp_directory

    args.pad_target_edges = not args.no_edge_padding
    args.pad_fourier = not args.no_fourier_padding

    if args.score not in MATCHING_EXHAUSTIVE_REGISTER:
        raise ValueError(
            f"score has to be one of {', '.join(MATCHING_EXHAUSTIVE_REGISTER.keys())}"
        )

    gpu_devices = os.environ.get("CUDA_VISIBLE_DEVICES", None)
    if args.gpu_indices is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu_indices

    if args.use_gpu:
        gpu_devices = os.environ.get("CUDA_VISIBLE_DEVICES", None)
        if gpu_devices is None:
            print(
                "No GPU indices provided and CUDA_VISIBLE_DEVICES is not set.",
                "Assuming device 0.",
            )
            os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        args.gpu_indices = [
            int(x) for x in os.environ["CUDA_VISIBLE_DEVICES"].split(",")
        ]

    return args


def main():
    args = parse_args()
    print_entry()

    target = Density.from_file(args.target, use_memmap=True)

    try:
        template = Density.from_file(args.template)
    except Exception:
        template = Density.from_structure(
            filename_or_structure=args.template,
            sampling_rate=target.sampling_rate,
        )

    if not np.allclose(target.sampling_rate, template.sampling_rate):
        print(
            f"Resampling template to {target.sampling_rate}. "
            "Consider providing a template with the same sampling rate as the target."
        )
        template = template.resample(target.sampling_rate, order=3)

    template_mask = load_and_validate_mask(
        mask_target=template, mask_path=args.template_mask
    )
    target_mask = load_and_validate_mask(
        mask_target=target, mask_path=args.target_mask, use_memmap=True
    )

    initial_shape = target.shape
    is_cropped = crop_data(
        data=target, data_mask=target_mask, cutoff=args.cutoff_target
    )
    print_block(
        name="Target",
        data={
            "Initial Shape": initial_shape,
            "Sampling Rate": tuple(np.round(target.sampling_rate, 2)),
            "Final Shape": target.shape,
        },
    )
    if is_cropped:
        args.target = generate_tempfile_name(suffix=".mrc")
        target.to_file(args.target)

        if target_mask:
            args.target_mask = generate_tempfile_name(suffix=".mrc")
            target_mask.to_file(args.target_mask)
            print_block(
                name="Target Mask",
                data={
                    "Initial Shape": initial_shape,
                    "Sampling Rate": tuple(np.round(target_mask.sampling_rate, 2)),
                    "Final Shape": target_mask.shape,
                },
            )

    initial_shape = template.shape
    _ = crop_data(data=template, data_mask=template_mask, cutoff=args.cutoff_template)
    template, translation = template.centered(0)
    print_block(
        name="Template",
        data={
            "Initial Shape": initial_shape,
            "Sampling Rate": tuple(np.round(template.sampling_rate, 2)),
            "Final Shape": template.shape,
        },
    )

    template_filter = {}
    if args.gaussian_sigma is not None:
        template.data = Preprocessor().gaussian_filter(
            sigma=args.gaussian_sigma, template=template.data
        )

    if args.bandpass_band is not None:
        bandpass_start, bandpass_stop = [
            float(x) for x in args.bandpass_band.split(",")
        ]
        if args.bandpass_smooth is None:
            args.bandpass_smooth = 0

        template_filter["bandpass_mask"] = {
            "minimum_frequency": bandpass_start,
            "maximum_frequency": bandpass_stop,
            "gaussian_sigma": args.bandpass_smooth,
        }

    if args.tilt_range is not None:
        args.wedge_smooth if args.wedge_smooth is not None else 0
        tilt_start, tilt_stop = [float(x) for x in args.tilt_range.split(",")]

        if args.tilt_step is not None:
            tilt_angles = np.arange(
                -tilt_start, tilt_stop + args.tilt_step, args.tilt_step
            )
            angles = np.zeros((template.data.ndim, tilt_angles.size))
            angles[2, :] = tilt_angles
            template_filter["wedge_mask"] = {
                "tilt_angles": angles,
                "sigma": args.wedge_smooth,
            }
        else:
            template_filter["continuous_wedge_mask"] = {
                "start_tilt": tilt_start,
                "stop_tilt": tilt_stop,
                "tilt_axis": 1,
                "infinite_plane": True,
                "sigma": args.wedge_smooth,
            }

    if template_mask is None:
        enclosing_box = template.minimum_enclosing_box(0, use_geometric_center=False)
        template_mask = template.empty
        template_mask.adjust_box(enclosing_box)
        template_mask.data[:] = 1
        translation = np.zeros_like(translation)

    template_mask.pad(template.shape, center=False)
    origin_translation = np.divide(
        np.subtract(template.origin, template_mask.origin), template.sampling_rate
    )
    translation = np.add(translation, origin_translation)

    template_mask = template_mask.rigid_transform(
        rotation_matrix=np.eye(template_mask.data.ndim),
        translation=-translation,
    )

    print_block(
        name="Template Mask",
        data={
            "Inital Shape": initial_shape,
            "Sampling Rate": tuple(np.round(template_mask.sampling_rate, 2)),
            "Final Shape": template_mask.shape,
        },
    )
    print("\n" + "-" * 80)

    if args.scramble_phases:
        template.data = scramble_phases(template.data, noise_proportion=1.0)

    available_memory = backend.get_available_memory()
    if args.use_gpu:
        args.cores = len(args.gpu_indices)
        has_torch = importlib.util.find_spec("torch") is not None
        has_cupy = importlib.util.find_spec("cupy") is not None

        if not has_torch and not has_cupy:
            raise ValueError(
                "Found neither CuPy nor PyTorch installation. You need to install"
                " either to enable GPU support."
            )

        if args.peak_calling:
            preferred_backend = "pytorch"
            if not has_torch:
                preferred_backend = "cupy"
            backend.change_backend(backend_name=preferred_backend, device="cuda")
        else:
            preferred_backend = "cupy"
            if not has_cupy:
                preferred_backend = "pytorch"
            backend.change_backend(backend_name=preferred_backend, device="cuda")
            if args.use_mixed_precision and preferred_backend == "pytorch":
                raise NotImplementedError(
                    "pytorch backend does not yet support mixed precision."
                    " Consider installing CuPy to enable this feature."
                )
            elif args.use_mixed_precision:
                backend.change_backend(
                    backend_name="cupy",
                    default_dtype=backend._array_backend.float16,
                    complex_dtype=backend._array_backend.complex64,
                    default_dtype_int=backend._array_backend.int16,
                )
        available_memory = backend.get_available_memory() * args.cores

    if args.memory is None:
        args.memory = int(args.memory_scaling * available_memory)

    target_padding = np.zeros_like(template.shape)
    if args.pad_target_edges:
        target_padding = template.shape

    template_box = template.shape
    if not args.pad_fourier:
        template_box = np.ones(len(template_box), dtype=int)

    callback_class = MaxScoreOverRotations
    if args.peak_calling:
        callback_class = PeakCallerMaximumFilter

    splits, schedule = compute_parallelization_schedule(
        shape1=target.shape,
        shape2=template_box,
        shape1_padding=target_padding,
        max_cores=args.cores,
        max_ram=args.memory,
        split_only_outer=args.use_gpu,
        matching_method=args.score,
        analyzer_method=callback_class.__name__,
        backend=backend._backend_name,
        float_nbytes=backend.datatype_bytes(backend._default_dtype),
        complex_nbytes=backend.datatype_bytes(backend._complex_dtype),
        integer_nbytes=backend.datatype_bytes(backend._default_dtype_int),
    )

    if splits is None:
        print(
            "Found no suitable parallelization schedule. Consider increasing"
            " available RAM or decreasing number of cores."
        )
        exit(-1)

    analyzer_args = {
        "score_threshold": 0.0,
        "number_of_peaks": 1000,
        "convolution_mode": "valid",
        "use_memmap": args.use_memmap,
    }

    matching_setup, matching_score = MATCHING_EXHAUSTIVE_REGISTER[args.score]
    matching_data = MatchingData(target=target, template=template.data)
    matching_data.rotations = get_rotation_matrices(
        angular_sampling=args.angular_sampling, dim=target.data.ndim
    )

    matching_data.template_filter = template_filter
    matching_data._invert_target = args.invert_target_contrast
    if target_mask is not None:
        matching_data.target_mask = target_mask
    if template_mask is not None:
        matching_data.template_mask = template_mask.data

    n_splits = np.prod(list(splits.values()))
    target_split = ", ".join(
        [":".join([str(x) for x in axis]) for axis in splits.items()]
    )
    gpus_used = 0 if args.gpu_indices is None else len(args.gpu_indices)
    options = {
        "CPU Cores": args.cores,
        "Run on GPU": f"{args.use_gpu} [N={gpus_used}]",
        "Use Mixed Precision": args.use_mixed_precision,
        "Assigned Memory [MB]": f"{args.memory // 1e6} [out of {available_memory//1e6}]",
        "Temporary Directory": args.temp_directory,
        "Extend Fourier Grid": not args.no_fourier_padding,
        "Extend Target Edges": not args.no_edge_padding,
        "Interpolation Order": args.interpolation_order,
        "Score": f"{args.score}",
        "Setup Function": f"{get_func_fullname(matching_setup)}",
        "Scoring Function": f"{get_func_fullname(matching_score)}",
        "Angular Sampling": f"{args.angular_sampling}"
        f" [{matching_data.rotations.shape[0]} rotations]",
        "Scramble Template": args.scramble_phases,
        "Target Splits": f"{target_split} [N={n_splits}]",
    }

    print_block(
        name="Template Matching Options",
        data=options,
        label_width=max(len(key) for key in options.keys()) + 2,
    )

    options = {"Analyzer": callback_class, **analyzer_args}
    print_block(
        name="Score Analysis Options",
        data=options,
        label_width=max(len(key) for key in options.keys()) + 2,
    )
    print("\n" + "-" * 80)

    outer_jobs = f"{schedule[0]} job{'s' if schedule[0] > 1 else ''}"
    inner_jobs = f"{schedule[1]} core{'s' if schedule[1] > 1 else ''}"
    n_splits = f"{n_splits} split{'s' if n_splits > 1 else ''}"
    print(f"\nDistributing {n_splits} on {outer_jobs} each using {inner_jobs}.")

    start = time()
    print("Running Template Matching. This might take a while ...")
    candidates = scan_subsets(
        matching_data=matching_data,
        job_schedule=schedule,
        matching_score=matching_score,
        matching_setup=matching_setup,
        callback_class=callback_class,
        callback_class_args=analyzer_args,
        target_splits=splits,
        pad_target_edges=args.pad_target_edges,
        pad_fourier=args.pad_fourier,
        interpolation_order=args.interpolation_order,
    )

    candidates = list(candidates) if candidates is not None else []
    if callback_class == MaxScoreOverRotations:
        if target_mask is not None and args.score != "MCC":
            candidates[0] *= target_mask.data
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            candidates[3] = {
                x: euler_from_rotationmatrix(
                    np.frombuffer(i, dtype=matching_data.rotations.dtype).reshape(
                        candidates[0].ndim, candidates[0].ndim
                    )
                )
                for i, x in candidates[3].items()
            }

    candidates.append((target.origin, template.origin, target.sampling_rate, args))
    write_pickle(data=candidates, filename=args.output)

    runtime = time() - start
    print(f"\nRuntime real: {runtime:.3f}s user: {(runtime * args.cores):.3f}s.")


if __name__ == "__main__":
    main()



#!python3
""" Estimate RAM requirements for template matching jobs.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import numpy as np
import argparse
from tme import Density
from tme.matching_utils import estimate_ram_usage


def parse_args():
    parser = argparse.ArgumentParser(
        description="Estimate RAM usage for template matching."
    )
    parser.add_argument(
        "-m",
        "--target",
        dest="target",
        type=str,
        required=True,
        help="Path to a target in CCP4/MRC format.",
    )
    parser.add_argument(
        "-i",
        "--template",
        dest="template",
        type=str,
        required=True,
        help="Path to a template in PDB/MMCIF or CCP4/MRC format.",
    )
    parser.add_argument(
        "--matching_method",
        required=False,
        default=None,
        help="Analyzer method to use.",
    )
    parser.add_argument(
        "--analyzer_method",
        required=False,
        default=None,
        help="Analyzer method to use.",
    )
    parser.add_argument(
        "--ncores", type=int, help="Number of cores for parallelization.", required=True
    )
    parser.add_argument(
        "--no_edge_padding",
        dest="no_edge_padding",
        action="store_true",
        default=False,
        help="Whether to pad the edges of the target. This is useful, if the target"
        " has a well defined bounding box, e.g. a density map.",
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    target = Density.from_file(args.target)
    template = Density.from_file(args.template)

    target_box = target.shape
    if not args.no_edge_padding:
        target_box = np.add(target_box, template.shape)

    result = estimate_ram_usage(
        shape1=target_box,
        shape2=template.shape,
        matching_method=args.matching_method,
        ncores=args.ncores,
        analyzer_method=args.analyzer_method,
    )
    print(result)


if __name__ == "__main__":
    main()
